"""
Tests for MCP server functionality.

Tests the refactored MCP server that uses the service layer
(RepositoryService, TagService, EventService).
"""

import pytest
import json
from unittest.mock import patch, MagicMock, PropertyMock
from datetime import datetime

from repoindex.domain import Repository, Tag, TagSource, Event
from repoindex.domain.repository import GitStatus, PackageMetadata


class TestMCPServer:
    """Test MCP server core functionality."""

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_create_server(self, mock_ctx_create):
        """Test creating MCP server."""
        # Mock context
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()

        # Should have resources
        assert len(server.resources) > 0
        assert 'repo://list' in server.resources

        # Should have tools
        assert len(server.tools) > 0
        assert 'repoindex_tag' in server.tools

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_list_resources(self, mock_ctx_create):
        """Test listing resources."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        resources = server.list_resources()

        # Should return list of resource definitions
        assert isinstance(resources, list)
        assert len(resources) > 0

        # Each resource should have required fields
        for resource in resources:
            assert 'uri' in resource
            assert 'name' in resource
            assert 'description' in resource

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_list_tools(self, mock_ctx_create):
        """Test listing tools."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        tools = server.list_tools()

        # Should return list of tool definitions
        assert isinstance(tools, list)
        assert len(tools) > 0

        # Each tool should have required fields
        for tool in tools:
            assert 'name' in tool
            assert 'description' in tool
            assert 'inputSchema' in tool


class TestMCPResources:
    """Test MCP resource handlers."""

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_read_repo_list(self, mock_ctx_create):
        """Test reading repo://list resource."""
        # Create mock repositories
        repos = [
            Repository(path='/tmp/repos/project1', name='project1', language='Python'),
            Repository(path='/tmp/repos/project2', name='project2', language='Rust'),
        ]

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = repos
        mock_ctx.tag_service.get_tag_strings.return_value = ['topic:ml']
        mock_ctx.tag_service.get_unique_tags.return_value = {'topic:ml'}
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.read_resource('repo://list')

        assert 'repositories' in result
        assert 'count' in result
        assert result['count'] == 2

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_read_tags_list(self, mock_ctx_create):
        """Test reading tags://list resource."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = {'topic:ml', 'work/active'}
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.read_resource('tags://list')

        assert 'tags' in result
        assert 'count' in result
        assert result['count'] == 2

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_read_stats_summary(self, mock_ctx_create):
        """Test reading stats://summary resource."""
        repos = [
            Repository(
                path='/tmp/repos/project1',
                name='project1',
                language='Python',
                package=PackageMetadata(name='project1', version='1.0.0', published=True, registry='pypi')
            ),
            Repository(path='/tmp/repos/project2', name='project2', language='Python'),
        ]

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = repos
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.read_resource('stats://summary')

        assert 'total_repositories' in result
        assert 'languages' in result
        assert 'published_packages' in result
        assert result['total_repositories'] == 2
        assert result['published_packages'] == 1


class TestMCPTools:
    """Test MCP tool handlers."""

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_tag(self, mock_ctx_create):
        """Test calling repoindex_tag tool."""
        repo = Repository(path='/tmp/repos/myproject', name='myproject')

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = [repo]
        mock_ctx.find_repo_by_name.return_value = repo
        mock_ctx.tag_service.get_explicit_tags.return_value = set()
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_tag', {'repo': 'myproject', 'tag': 'topic:ml'})

        assert result['success'] is True
        assert result['action'] == 'added'
        mock_ctx.tag_service.add.assert_called_once()

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_tag_already_exists(self, mock_ctx_create):
        """Test calling repoindex_tag when tag already exists."""
        repo = Repository(path='/tmp/repos/myproject', name='myproject')
        existing_tag = Tag.parse('topic:ml', TagSource.EXPLICIT)

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = [repo]
        mock_ctx.find_repo_by_name.return_value = repo
        mock_ctx.tag_service.get_explicit_tags.return_value = {existing_tag}
        mock_ctx.tag_service.get_unique_tags.return_value = {'topic:ml'}
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_tag', {'repo': 'myproject', 'tag': 'topic:ml'})

        assert result['success'] is True
        assert result['action'] == 'already_exists'

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_untag(self, mock_ctx_create):
        """Test calling repoindex_untag tool."""
        repo = Repository(path='/tmp/repos/myproject', name='myproject')

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = [repo]
        mock_ctx.find_repo_by_name.return_value = repo
        mock_ctx.tag_service.remove.return_value = True
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_untag', {'repo': 'myproject', 'tag': 'topic:ml'})

        assert result['success'] is True
        assert result['action'] == 'removed'

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_stats(self, mock_ctx_create):
        """Test calling repoindex_stats tool."""
        repos = [
            Repository(path='/tmp/repos/project1', name='project1', language='Python'),
            Repository(path='/tmp/repos/project2', name='project2', language='Python'),
        ]

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = repos
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_stats', {'groupby': 'language'})

        assert 'groupby' in result
        assert 'groups' in result
        assert result['groupby'] == 'language'
        assert result['groups']['Python'] == 2

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_refresh(self, mock_ctx_create):
        """Test calling repoindex_refresh tool."""
        repos = [
            Repository(path='/tmp/repos/project1', name='project1'),
            Repository(path='/tmp/repos/project2', name='project2'),
        ]

        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = repos
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_refresh', {})

        assert 'refreshed' in result
        assert 'count' in result
        assert result['count'] == 2

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_unknown_tool(self, mock_ctx_create):
        """Test calling unknown tool raises error."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()

        with pytest.raises(ValueError, match="Unknown tool"):
            server.call_tool('unknown_tool', {})

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_call_repoindex_tag_repo_not_found(self, mock_ctx_create):
        """Test calling repoindex_tag when repo not found."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.find_repo_by_name.return_value = None
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()
        result = server.call_tool('repoindex_tag', {'repo': 'nonexistent', 'tag': 'topic:ml'})

        assert result['success'] is False
        assert 'error' in result


class TestMCPJSONRPC:
    """Test JSON-RPC request handling."""

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_handle_initialize(self, mock_ctx_create):
        """Test handling initialize request."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server, _handle_jsonrpc_request

        server = create_mcp_server()
        request = {
            "jsonrpc": "2.0",
            "method": "initialize",
            "id": 1
        }

        response = _handle_jsonrpc_request(server, request)

        assert response['jsonrpc'] == '2.0'
        assert response['id'] == 1
        assert 'result' in response
        assert 'serverInfo' in response['result']
        assert response['result']['serverInfo']['name'] == 'repoindex'

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_handle_resources_list(self, mock_ctx_create):
        """Test handling resources/list request."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server, _handle_jsonrpc_request

        server = create_mcp_server()
        request = {
            "jsonrpc": "2.0",
            "method": "resources/list",
            "id": 2
        }

        response = _handle_jsonrpc_request(server, request)

        assert 'result' in response
        assert 'resources' in response['result']
        assert len(response['result']['resources']) > 0

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_handle_tools_list(self, mock_ctx_create):
        """Test handling tools/list request."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server, _handle_jsonrpc_request

        server = create_mcp_server()
        request = {
            "jsonrpc": "2.0",
            "method": "tools/list",
            "id": 3
        }

        response = _handle_jsonrpc_request(server, request)

        assert 'result' in response
        assert 'tools' in response['result']
        assert len(response['result']['tools']) > 0

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_handle_unknown_method(self, mock_ctx_create):
        """Test handling unknown method returns error."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server, _handle_jsonrpc_request

        server = create_mcp_server()
        request = {
            "jsonrpc": "2.0",
            "method": "unknown/method",
            "id": 4
        }

        response = _handle_jsonrpc_request(server, request)

        assert 'error' in response
        assert response['error']['code'] == -32601


class TestURIPatternMatching:
    """Test URI pattern matching for resources."""

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_match_simple_uri(self, mock_ctx_create):
        """Test matching simple URI pattern."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()

        # Should match exact pattern
        assert server._match_uri_pattern('repo://list', 'repo://list')

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_match_parameterized_uri(self, mock_ctx_create):
        """Test matching URI with parameters."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()

        # Should match parameterized pattern
        assert server._match_uri_pattern('repo://{name}', 'repo://myproject')
        assert server._match_uri_pattern('repo://{name}/status', 'repo://myproject/status')

    @patch('repoindex.mcp.server.MCPContext.create')
    def test_extract_uri_params(self, mock_ctx_create):
        """Test extracting parameters from URI."""
        mock_ctx = MagicMock()
        mock_ctx.discover_repos.return_value = []
        mock_ctx.tag_service.get_unique_tags.return_value = set()
        mock_ctx_create.return_value = mock_ctx

        from repoindex.mcp.server import create_mcp_server

        server = create_mcp_server()

        params = server._extract_uri_params('repo://{name}', 'repo://myproject')
        assert params == {'name': 'myproject'}

        params = server._extract_uri_params('tags://{tag}/repos', 'tags://topic:ml/repos')
        assert params == {'tag': 'topic:ml'}


class TestMCPContext:
    """Test MCPContext functionality."""

    @patch('repoindex.mcp.server.load_config')
    @patch('repoindex.mcp.server.FileStore')
    @patch('repoindex.mcp.server.GitClient')
    @patch('repoindex.mcp.server.GitHubClient')
    @patch('repoindex.mcp.server.RepositoryService')
    @patch('repoindex.mcp.server.TagService')
    @patch('repoindex.mcp.server.EventService')
    def test_create_context(self, mock_event_svc, mock_tag_svc, mock_repo_svc,
                           mock_gh_client, mock_git_client, mock_file_store,
                           mock_load_config):
        """Test creating MCPContext."""
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']}
        }

        from repoindex.mcp.server import MCPContext

        ctx = MCPContext.create()

        assert ctx.config is not None
        assert ctx.repo_service is not None
        assert ctx.tag_service is not None
        assert ctx.event_service is not None


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



"""
GitHub API client infrastructure for repoindex.

Provides a clean abstraction over GitHub API access:
- Uses `gh` CLI when available for authentication
- Falls back to requests with token
- Handles rate limiting with exponential backoff
"""

import subprocess
import json
import os
import time
import logging
from dataclasses import dataclass
from typing import Optional, List, Dict, Any
from datetime import datetime

logger = logging.getLogger(__name__)


@dataclass
class GitHubRepo:
    """GitHub repository metadata."""
    owner: str
    name: str
    full_name: str
    description: Optional[str]
    homepage: Optional[str]
    language: Optional[str]
    stars: int
    forks: int
    watchers: int
    open_issues: int
    is_fork: bool
    is_private: bool
    is_archived: bool
    default_branch: str
    topics: List[str]
    license_key: Optional[str]
    has_issues: bool
    has_wiki: bool
    has_pages: bool
    created_at: Optional[str]
    updated_at: Optional[str]
    pushed_at: Optional[str]

    @classmethod
    def from_api_response(cls, data: Dict[str, Any]) -> 'GitHubRepo':
        """Create from GitHub API response."""
        owner = data.get('owner', {})
        license_info = data.get('license', {})

        return cls(
            owner=owner.get('login', '') if isinstance(owner, dict) else str(owner),
            name=data.get('name', ''),
            full_name=data.get('full_name', ''),
            description=data.get('description'),
            homepage=data.get('homepage'),
            language=data.get('language'),
            stars=data.get('stargazers_count', 0),
            forks=data.get('forks_count', 0),
            watchers=data.get('watchers_count', 0),
            open_issues=data.get('open_issues_count', 0),
            is_fork=data.get('fork', False),
            is_private=data.get('private', False),
            is_archived=data.get('archived', False),
            default_branch=data.get('default_branch', 'main'),
            topics=data.get('topics', []),
            license_key=license_info.get('key') if isinstance(license_info, dict) else None,
            has_issues=data.get('has_issues', True),
            has_wiki=data.get('has_wiki', True),
            has_pages=data.get('has_pages', False),
            created_at=data.get('created_at'),
            updated_at=data.get('updated_at'),
            pushed_at=data.get('pushed_at'),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            'owner': self.owner,
            'name': self.name,
            'full_name': self.full_name,
            'description': self.description,
            'homepage': self.homepage,
            'language': self.language,
            'stars': self.stars,
            'forks': self.forks,
            'watchers': self.watchers,
            'open_issues': self.open_issues,
            'is_fork': self.is_fork,
            'is_private': self.is_private,
            'is_archived': self.is_archived,
            'default_branch': self.default_branch,
            'topics': self.topics,
            'license_key': self.license_key,
            'has_issues': self.has_issues,
            'has_wiki': self.has_wiki,
            'has_pages': self.has_pages,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'pushed_at': self.pushed_at,
        }


class GitHubClient:
    """
    GitHub API client with rate limiting.

    Uses `gh` CLI for authentication when available,
    with fallback to direct API calls with token.

    Example:
        client = GitHubClient()
        repo = client.get_repo("owner", "repo")
        if repo:
            print(f"Stars: {repo.stars}")
    """

    def __init__(
        self,
        token: Optional[str] = None,
        max_retries: int = 3,
        base_delay: float = 1.0,
        max_delay: float = 60.0
    ):
        """
        Initialize GitHubClient.

        Args:
            token: GitHub token (defaults to REPOINDEX_GITHUB_TOKEN or GITHUB_TOKEN env var)
            max_retries: Maximum retry attempts for rate-limited requests
            base_delay: Base delay for exponential backoff
            max_delay: Maximum delay between retries
        """
        self.token = token or os.environ.get('REPOINDEX_GITHUB_TOKEN') or os.environ.get('GITHUB_TOKEN')
        self.max_retries = max_retries
        self.base_delay = base_delay
        self.max_delay = max_delay
        self._use_gh_cli = self._check_gh_cli()

    def _check_gh_cli(self) -> bool:
        """Check if gh CLI is available and authenticated."""
        try:
            result = subprocess.run(
                ['gh', 'auth', 'status'],
                capture_output=True,
                text=True,
                timeout=10
            )
            return result.returncode == 0
        except (subprocess.TimeoutExpired, FileNotFoundError):
            return False

    def _gh_api(self, endpoint: str) -> Optional[Dict[str, Any]]:
        """Call GitHub API using gh CLI."""
        try:
            result = subprocess.run(
                ['gh', 'api', endpoint],
                capture_output=True,
                text=True,
                timeout=30
            )
            if result.returncode == 0 and result.stdout:
                return json.loads(result.stdout)
            return None
        except (subprocess.TimeoutExpired, json.JSONDecodeError, FileNotFoundError) as e:
            logger.debug(f"gh api call failed for {endpoint}: {e}")
            return None

    def _requests_api(self, endpoint: str) -> Optional[Dict[str, Any]]:
        """Call GitHub API using requests library."""
        try:
            import requests
        except ImportError:
            logger.warning("requests library not available for GitHub API")
            return None

        url = f"https://api.github.com/{endpoint}"
        headers = {
            'Accept': 'application/vnd.github.v3+json',
            'User-Agent': 'repoindex'
        }

        if self.token:
            headers['Authorization'] = f'token {self.token}'

        for attempt in range(self.max_retries):
            try:
                response = requests.get(url, headers=headers, timeout=30)

                if response.status_code == 200:
                    return response.json()

                if response.status_code == 403:
                    # Rate limited
                    reset_time = response.headers.get('X-RateLimit-Reset')
                    if reset_time:
                        wait_time = int(reset_time) - int(time.time())
                        if wait_time > 0 and wait_time < self.max_delay:
                            logger.info(f"Rate limited, waiting {wait_time}s")
                            time.sleep(wait_time)
                            continue

                    # Exponential backoff
                    delay = min(self.base_delay * (2 ** attempt), self.max_delay)
                    logger.info(f"Rate limited, waiting {delay}s (attempt {attempt + 1})")
                    time.sleep(delay)
                    continue

                if response.status_code == 404:
                    return None

                logger.warning(f"GitHub API error {response.status_code} for {endpoint}")
                return None

            except requests.RequestException as e:
                logger.warning(f"GitHub API request failed: {e}")
                if attempt < self.max_retries - 1:
                    delay = min(self.base_delay * (2 ** attempt), self.max_delay)
                    time.sleep(delay)
                continue

        return None

    def _api(self, endpoint: str) -> Optional[Dict[str, Any]]:
        """Call GitHub API using best available method."""
        if self._use_gh_cli:
            result = self._gh_api(endpoint)
            if result is not None:
                return result

        return self._requests_api(endpoint)

    def get_repo(self, owner: str, name: str) -> Optional[GitHubRepo]:
        """
        Get repository metadata.

        Args:
            owner: Repository owner
            name: Repository name

        Returns:
            GitHubRepo or None if not found
        """
        data = self._api(f"repos/{owner}/{name}")
        if data:
            return GitHubRepo.from_api_response(data)
        return None

    def get_topics(self, owner: str, name: str) -> List[str]:
        """
        Get repository topics.

        Args:
            owner: Repository owner
            name: Repository name

        Returns:
            List of topic strings
        """
        data = self._api(f"repos/{owner}/{name}/topics")
        if data:
            return data.get('names', [])
        return []

    def repo_exists(self, owner: str, name: str) -> bool:
        """Check if repository exists and is accessible."""
        return self.get_repo(owner, name) is not None

    def get_releases(self, owner: str, name: str, limit: int = 10) -> List[Dict[str, Any]]:
        """
        Get repository releases.

        Args:
            owner: Repository owner
            name: Repository name
            limit: Maximum releases to return

        Returns:
            List of release data dictionaries
        """
        data = self._api(f"repos/{owner}/{name}/releases?per_page={limit}")
        if data and isinstance(data, list):
            return data[:limit]
        return []

    def get_pages_info(self, owner: str, name: str) -> Optional[Dict[str, Any]]:
        """
        Get GitHub Pages information.

        Args:
            owner: Repository owner
            name: Repository name

        Returns:
            Pages info dict or None if not enabled
        """
        return self._api(f"repos/{owner}/{name}/pages")



"""
MCP (Model Context Protocol) server for repoindex.

Exposes repoindex as an MCP server for integration with LLM tools like Claude Code.

Resources (read-only data):
    repo://list                 - All repositories with basic metadata
    repo://{name}               - Full metadata for one repository
    repo://{name}/status        - Git status for one repository
    repo://{name}/package       - Package info (PyPI/CRAN/npm)
    tags://list                 - All tags
    tags://tree                 - Hierarchical tag view
    tags://{tag}/repos          - Repositories with this tag
    stats://summary             - Overall statistics
    stats://languages           - Count by language
    stats://published           - Registry publication status
    events://recent             - Recent events
    events://repo/{name}        - Events for one repository
    events://type/{type}        - Events by type

Tools (actions):
    repoindex_tag(repo, tag)        - Add tag to repository
    repoindex_untag(repo, tag)      - Remove tag from repository
    repoindex_query(expression)     - Query repositories
    repoindex_refresh(repo?)        - Refresh metadata
    repoindex_stats(groupby)        - Get statistics
"""

from .server import create_mcp_server, run_mcp_server

__all__ = ['create_mcp_server', 'run_mcp_server']



"""
Status dashboard command for repoindex.

Shows a quick overview of your repository collection from the database.
"""

import json
import sys
from datetime import datetime, timedelta

import click

from ..config import load_config
from ..database import Database, get_database_info


@click.command(name='status')
@click.option('--json', 'output_json', is_flag=True, help='Output as JSON for scripting')
@click.option('--repos', is_flag=True, help='List individual repositories')
@click.option('--pretty/--no-pretty', default=True, help='Pretty format (default: auto-detect)')
def status_handler(output_json: bool, repos: bool, pretty: bool):
    """
    Show repository collection status dashboard.

    Displays a quick overview of your repository collection from the database.

    Examples:

        # Dashboard view
        repoindex status

        # JSON output for scripting
        repoindex status --json

        # List all repos
        repoindex status --repos
    """
    config = load_config()

    # Check if database exists
    try:
        db_info = get_database_info(config)
    except Exception:
        click.echo("Database not found. Run 'repoindex refresh' first.", err=True)
        sys.exit(1)

    if repos:
        _list_repos(config, output_json, pretty)
        return

    if output_json:
        _json_dashboard(config, db_info)
    else:
        _pretty_dashboard(config, db_info)


def _json_dashboard(config: dict, db_info: dict):
    """Output dashboard data as JSON."""
    data = _gather_dashboard_data(config, db_info)
    print(json.dumps(data, indent=2, default=str))


def _pretty_dashboard(config: dict, db_info: dict):
    """Display a pretty dashboard."""
    from rich.console import Console
    from rich.table import Table
    from rich.panel import Panel
    from rich import box

    console = Console()
    data = _gather_dashboard_data(config, db_info)

    # Title
    console.print()
    console.print("[bold cyan]repoindex Status Dashboard[/bold cyan]")
    console.print("=" * 50)
    console.print()

    # Database section
    db = data['database']
    size_str = _format_size(db.get('size_bytes', 0))
    console.print(f"[bold]Database:[/bold]    {db['repos']} repos, {db['events']} events ({size_str})")

    if db.get('last_refresh'):
        refresh_ago = _format_time_ago(db['last_refresh'])
        console.print(f"[bold]Last refresh:[/bold] {refresh_ago}")
    else:
        console.print("[bold]Last refresh:[/bold] [yellow]never[/yellow]")

    console.print()

    # Health section
    health = data['health']
    total = health['clean'] + health['dirty']
    if total > 0:
        clean_pct = health['clean'] * 100 // total
        dirty_pct = 100 - clean_pct

        console.print("[bold]Health:[/bold]")
        console.print(f"  [green]Clean:[/green]     {health['clean']} repos ({clean_pct}%)")
        if health['dirty'] > 0:
            console.print(f"  [yellow]Dirty:[/yellow]     {health['dirty']} repos ({dirty_pct}%)")
        console.print()

    # Languages section
    if data['languages']:
        lang_parts = [f"{lang} ({count})" for lang, count in data['languages'][:5]]
        if len(data['languages']) > 5:
            lang_parts.append(f"+{len(data['languages']) - 5} more")
        console.print(f"[bold]Languages:[/bold]   {', '.join(lang_parts)}")
        console.print()

    # Recent activity section
    activity = data['recent_activity']
    if activity['commits'] > 0 or activity['tags'] > 0:
        console.print(f"[bold]Recent Activity ({activity['since']}):[/bold]")
        if activity['commits'] > 0:
            console.print(f"  Commits:    {activity['commits']} across {activity['active_repos']} repos")
        if activity['tags'] > 0:
            console.print(f"  Tags:       {activity['tags']} new releases")
        console.print()

    # Footer
    if not db.get('last_refresh'):
        console.print("[dim]Run 'repoindex refresh' to populate the database.[/dim]")
    elif db['repos'] == 0:
        console.print("[dim]No repositories indexed. Check your configuration with 'repoindex config show'.[/dim]")
    else:
        console.print("[dim]Run 'repoindex refresh' to update, 'repoindex query' to search.[/dim]")


def _gather_dashboard_data(config: dict, db_info: dict) -> dict:
    """Gather all dashboard data from the database."""
    data = {
        'database': {
            'repos': db_info.get('repos', 0),
            'events': db_info.get('events', 0),
            'tags': db_info.get('tags', 0),
            'size_bytes': db_info.get('size_bytes', 0),
            'last_refresh': None,
        },
        'health': {
            'clean': 0,
            'dirty': 0,
        },
        'languages': [],
        'recent_activity': {
            'since': '7 days',
            'commits': 0,
            'tags': 0,
            'active_repos': 0,
        },
    }

    try:
        with Database(config=config, read_only=True) as db:
            # Get last refresh time
            db.execute("SELECT MAX(last_scan) as last_scan FROM repos")
            row = db.fetchone()
            if row and row['last_scan']:
                data['database']['last_refresh'] = row['last_scan']

            # Get health stats
            db.execute("SELECT COUNT(*) as count FROM repos WHERE is_clean = 1")
            row = db.fetchone()
            data['health']['clean'] = row['count'] if row else 0

            db.execute("SELECT COUNT(*) as count FROM repos WHERE is_clean = 0")
            row = db.fetchone()
            data['health']['dirty'] = row['count'] if row else 0

            # Get language breakdown
            db.execute("""
                SELECT language, COUNT(*) as count
                FROM repos
                WHERE language IS NOT NULL AND language != ''
                GROUP BY language
                ORDER BY count DESC
                LIMIT 10
            """)
            data['languages'] = [(row['language'], row['count']) for row in db.fetchall()]

            # Get recent activity (last 7 days)
            since = datetime.now() - timedelta(days=7)
            since_str = since.strftime('%Y-%m-%d')

            db.execute("""
                SELECT COUNT(*) as count FROM events
                WHERE type = 'commit' AND timestamp >= ?
            """, (since_str,))
            row = db.fetchone()
            data['recent_activity']['commits'] = row['count'] if row else 0

            db.execute("""
                SELECT COUNT(*) as count FROM events
                WHERE type = 'git_tag' AND timestamp >= ?
            """, (since_str,))
            row = db.fetchone()
            data['recent_activity']['tags'] = row['count'] if row else 0

            db.execute("""
                SELECT COUNT(DISTINCT repo_id) as count FROM events
                WHERE type = 'commit' AND timestamp >= ?
            """, (since_str,))
            row = db.fetchone()
            data['recent_activity']['active_repos'] = row['count'] if row else 0

    except Exception:
        pass  # Database might not exist yet

    return data


def _list_repos(config: dict, output_json: bool, pretty: bool):
    """List individual repositories."""
    try:
        with Database(config=config, read_only=True) as db:
            db.execute("""
                SELECT name, path, language, branch, is_clean, stars
                FROM repos
                ORDER BY name
            """)
            rows = db.fetchall()

            if output_json:
                repos = [dict(row) for row in rows]
                print(json.dumps(repos, indent=2))
            elif pretty and sys.stdout.isatty():
                _print_repos_table(rows)
            else:
                for row in rows:
                    print(json.dumps(dict(row), default=str))

    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


def _print_repos_table(rows):
    """Print repositories as a table."""
    from rich.console import Console
    from rich.table import Table
    from rich import box

    console = Console()
    table = Table(title=f"Repositories ({len(rows)})", box=box.ROUNDED)

    table.add_column("Name", style="cyan")
    table.add_column("Language")
    table.add_column("Branch")
    table.add_column("Status")
    table.add_column("Stars", justify="right")

    for row in rows:
        status = "[green]clean[/green]" if row['is_clean'] else "[yellow]dirty[/yellow]"
        stars = str(row['stars']) if row['stars'] else ""
        table.add_row(
            row['name'] or "",
            row['language'] or "",
            row['branch'] or "",
            status,
            stars,
        )

    console.print(table)


def _format_size(bytes_size: int) -> str:
    """Format bytes as human readable."""
    for unit in ['B', 'KB', 'MB', 'GB']:
        if bytes_size < 1024:
            return f"{bytes_size:.0f} {unit}"
        bytes_size /= 1024
    return f"{bytes_size:.0f} TB"


def _format_time_ago(timestamp: str) -> str:
    """Format timestamp as time ago."""
    try:
        if isinstance(timestamp, str):
            dt = datetime.fromisoformat(timestamp.replace('Z', '+00:00'))
        else:
            dt = timestamp

        now = datetime.now(dt.tzinfo) if dt.tzinfo else datetime.now()
        delta = now - dt

        if delta.days > 30:
            return dt.strftime('%Y-%m-%d')
        elif delta.days > 0:
            return f"{delta.days} days ago"
        elif delta.seconds > 3600:
            hours = delta.seconds // 3600
            return f"{hours} hours ago"
        elif delta.seconds > 60:
            mins = delta.seconds // 60
            return f"{mins} minutes ago"
        else:
            return "just now"
    except Exception:
        return str(timestamp)



"""
Simplified TUI application - just the activity dashboard.
"""

from textual.app import App
from typing import Optional

from .dashboard import ActivityDashboard


class RepoIndexApp(App):
    """Simple htop-style activity monitor."""

    TITLE = "repoindex htop - Repository Activity Monitor"

    def __init__(self, config_path: Optional[str] = None):
        """Initialize app.

        Args:
            config_path: Optional path to config file
        """
        super().__init__()
        self.config_path = config_path

    def on_mount(self) -> None:
        """Mount the dashboard screen directly."""
        from ..config import load_config
        from ..utils import find_git_repos_from_config

        # Load config
        if self.config_path:
            import os
            os.environ['REPOINDEX_CONFIG'] = self.config_path

        config = load_config()

        # Find all repos
        repo_dirs = config.get('repository_directories', [])
        if not repo_dirs:
            repo_dirs = ['.']

        repos = find_git_repos_from_config(repo_dirs, recursive=False)

        # Push the activity dashboard as the only screen
        dashboard = ActivityDashboard(repos)
        self.push_screen(dashboard)


def run_tui(config_path: Optional[str] = None) -> None:
    """Run the htop application.

    Args:
        config_path: Path to configuration file
    """
    app = RepoIndexApp(config_path)
    app.run()



#!/usr/bin/env python3

import os
import json
import requests
import tomllib
from pathlib import Path
from typing import Dict, Optional, Tuple, List
import re

from .config import logger
from .config import load_config
from .utils import run_command

def find_packaging_files(repo_path: str) -> List[str]:
    """Find Python packaging files in a repository."""
    packaging_files = []
    repo_path = Path(repo_path)
    
    # Look for common packaging files
    candidates = ['pyproject.toml', 'setup.py', 'setup.cfg']
    
    for candidate in candidates:
        file_path = repo_path / candidate
        if file_path.exists():
            packaging_files.append(str(file_path))
    
    return packaging_files

def extract_package_name_from_pyproject(file_path: str) -> Optional[str]:
    """Extract package name from pyproject.toml."""
    try:
        with open(file_path, 'rb') as f:
            data = tomllib.load(f)
        
        # Check [project] section first (PEP 621)
        if 'project' in data and 'name' in data['project']:
            return data['project']['name']
        
        # Check [tool.setuptools] for older format
        if 'tool' in data and 'setuptools' in data['tool'] and 'name' in data['tool']['setuptools']:
            return data['tool']['setuptools']['name']
        
        # Check [build-system] for some edge cases
        if 'build-system' in data and 'name' in data['build-system']:
            return data['build-system']['name']
        
    except Exception as e:
        logger.debug(f"Error parsing {file_path}: {e}")
    
    return None


def extract_package_version_from_pyproject(file_path: str) -> Optional[str]:
    """Extract package version from pyproject.toml."""
    try:
        with open(file_path, 'rb') as f:
            data = tomllib.load(f)
        
        # Check [project] section first (PEP 621)
        if 'project' in data and 'version' in data['project']:
            return data['project']['version']
        
        # Check [tool.setuptools] for older format
        if 'tool' in data and 'setuptools' in data['tool'] and 'version' in data['tool']['setuptools']:
            return data['tool']['setuptools']['version']
        
    except Exception as e:
        logger.debug(f"Error parsing {file_path}: {e}")
    
    return None

def extract_package_name_from_setup_py(file_path: str) -> Optional[str]:
    """Extract package name from setup.py (basic regex parsing)."""
    try:
        with open(file_path, 'r') as f:
            content = f.read()
        
        # Look for name= parameter in setup() call
        patterns = [
            r'name\s*=\s*["\']([^"\']+)["\']',
            r'name\s*=\s*([a-zA-Z_][a-zA-Z0-9_-]*)',
        ]
        
        for pattern in patterns:
            match = re.search(pattern, content)
            if match:
                return match.group(1)
        
    except Exception as e:
        logger.debug(f"Error parsing {file_path}: {e}")
    
    return None

def extract_package_name_from_setup_cfg(file_path: str) -> Optional[str]:
    """Extract package name from setup.cfg."""
    try:
        import configparser
        config_parser = configparser.ConfigParser()
        config_parser.read(file_path)
        
        if 'metadata' in config_parser and 'name' in config_parser['metadata']:
            return config_parser['metadata']['name']
        
    except Exception as e:
        logger.debug(f"Error parsing {file_path}: {e}")
    
    return None

def extract_package_name(file_path: str) -> Optional[str]:
    """Extract package name from a packaging file."""
    file_path = Path(file_path)
    
    if file_path.name == 'pyproject.toml':
        return extract_package_name_from_pyproject(str(file_path))
    elif file_path.name == 'setup.py':
        return extract_package_name_from_setup_py(str(file_path))
    elif file_path.name == 'setup.cfg':
        return extract_package_name_from_setup_cfg(str(file_path))
    
    return None

def check_pypi_package(package_name: str) -> Optional[Dict]:
    """Check if a package exists on PyPI and get its info."""
    try:
        config = load_config()
        timeout = config.get('pypi', {}).get('timeout_seconds', 10)
        
        # Check main PyPI
        url = f"https://pypi.org/pypi/{package_name}/json"
        response = requests.get(url, timeout=timeout)
        
        if response.status_code == 200:
            data = response.json()
            result = {
                'exists': True,
                'version': data['info']['version'],
                'url': f"https://pypi.org/project/{package_name}/",
                'description': data['info']['summary'] or '',
                'author': data['info']['author'] or '',
                'home_page': data['info']['home_page'] or '',
                'download_url': data['info']['download_url'] or '',
                'last_updated': data['urls'][0]['upload_time'] if data['urls'] else ''
            }
            return result
        elif response.status_code == 404:
            return {'exists': False}
        else:
            logger.warning(f"PyPI API returned status {response.status_code} for {package_name}")
            return None
            
    except requests.RequestException as e:
        logger.debug(f"Error checking PyPI for {package_name}: {e}")
        return None
    except Exception as e:
        logger.debug(f"Unexpected error checking PyPI for {package_name}: {e}")
        return None

def detect_pypi_package(repo_path: str) -> Dict:
    """Detect PyPI package information for a repository."""
    result = {
        'has_packaging_files': False,
        'packaging_files': [],
        'package_name': None,
        'local_version': None,
        'pypi_info': None,
        'is_published': False
    }
    
    # Find packaging files
    packaging_files = find_packaging_files(repo_path)
    result['packaging_files'] = packaging_files
    result['has_packaging_files'] = bool(packaging_files)
    
    if not packaging_files:
        return result
    
    # Try to extract package name and version from packaging files
    for file_path in packaging_files:
        package_name = extract_package_name(file_path)
        if package_name:
            result['package_name'] = package_name
            # Also try to get version from the same file
            if 'pyproject.toml' in file_path:
                version = extract_package_version_from_pyproject(file_path)
                if version:
                    result['local_version'] = version
            break
    
    # Only check PyPI if we found a package name in the packaging files
    # Don't use directory name as fallback for PyPI checks
    if result['package_name']:
        pypi_info = check_pypi_package(result['package_name'])
        if pypi_info:
            result['pypi_info'] = pypi_info
            result['is_published'] = pypi_info.get('exists', False)
    
    return result

def get_local_package_version(repo_path: str, package_name: str) -> Optional[str]:
    """Get the local version of a package from packaging files."""
    packaging_files = find_packaging_files(repo_path)
    
    for file_path in packaging_files:
        if Path(file_path).name == 'pyproject.toml':
            try:
                with open(file_path, 'rb') as f:
                    data = tomllib.load(f)
                
                # Check [project] section
                if 'project' in data and 'version' in data['project']:
                    return data['project']['version']
                
                # Check [tool.setuptools] for older format
                if 'tool' in data and 'setuptools' in data['tool'] and 'version' in data['tool']['setuptools']:
                    return data['tool']['setuptools']['version']
                    
            except Exception as e:
                logger.debug(f"Error reading version from {file_path}: {e}")
        
        elif Path(file_path).name == 'setup.py':
            try:
                with open(file_path, 'r') as f:
                    content = f.read()
                
                # Look for version= parameter
                patterns = [
                    r'version\s*=\s*["\']([^"\']+)["\']',
                ]
                
                for pattern in patterns:
                    match = re.search(pattern, content)
                    if match:
                        return match.group(1)
                        
            except Exception as e:
                logger.debug(f"Error reading version from {file_path}: {e}")
    
    return None

def extract_classifiers_from_pyproject(file_path: str) -> List[str]:
    """Extract PyPI classifiers from pyproject.toml."""
    try:
        with open(file_path, 'rb') as f:
            data = tomllib.load(f)
        
        # Check [project] section (PEP 621)
        if 'project' in data and 'classifiers' in data['project']:
            return data['project']['classifiers']
        
        # Check [tool.setuptools] for older format
        if 'tool' in data and 'setuptools' in data['tool'] and 'classifiers' in data['tool']['setuptools']:
            return data['tool']['setuptools']['classifiers']
        
    except Exception as e:
        logger.debug(f"Error extracting classifiers from {file_path}: {e}")
    
    return []


def extract_classifiers_from_setup_py(file_path: str) -> List[str]:
    """Extract PyPI classifiers from setup.py (basic regex parsing)."""
    try:
        with open(file_path, 'r') as f:
            content = f.read()
        
        # Look for classifiers parameter in setup() call
        # This is a simplified approach - in reality, classifiers can be complex
        match = re.search(r'classifiers\s*=\s*\[(.*?)\]', content, re.DOTALL)
        if match:
            classifiers_text = match.group(1)
            # Extract strings from the list
            classifier_matches = re.findall(r'["\']([^"\']+)["\']', classifiers_text)
            return classifier_matches
        
    except Exception as e:
        logger.debug(f"Error extracting classifiers from {file_path}: {e}")
    
    return []


def extract_classifiers_from_setup_cfg(file_path: str) -> List[str]:
    """Extract PyPI classifiers from setup.cfg."""
    try:
        import configparser
        config_parser = configparser.ConfigParser()
        config_parser.read(file_path)
        
        if 'metadata' in config_parser and 'classifiers' in config_parser['metadata']:
            # Classifiers in setup.cfg are typically multiline
            classifiers_text = config_parser['metadata']['classifiers']
            return [line.strip() for line in classifiers_text.strip().split('\n') if line.strip()]
        
    except Exception as e:
        logger.debug(f"Error extracting classifiers from {file_path}: {e}")
    
    return []


def extract_keywords_from_packaging_files(repo_path: str) -> List[str]:
    """Extract keywords from packaging files."""
    keywords = []
    packaging_files = find_packaging_files(repo_path)
    
    for file_path in packaging_files:
        if Path(file_path).name == 'pyproject.toml':
            try:
                with open(file_path, 'rb') as f:
                    data = tomllib.load(f)
                
                # Check [project] section
                if 'project' in data and 'keywords' in data['project']:
                    keywords.extend(data['project']['keywords'])
                    
            except Exception as e:
                logger.debug(f"Error reading keywords from {file_path}: {e}")
        
        elif Path(file_path).name == 'setup.py':
            try:
                with open(file_path, 'r') as f:
                    content = f.read()
                
                # Look for keywords parameter
                match = re.search(r'keywords\s*=\s*["\']([^"\']+)["\']', content)
                if not match:
                    # Try list format
                    match = re.search(r'keywords\s*=\s*\[(.*?)\]', content, re.DOTALL)
                    if match:
                        keyword_text = match.group(1)
                        keyword_matches = re.findall(r'["\']([^"\']+)["\']', keyword_text)
                        keywords.extend(keyword_matches)
                else:
                    # Space or comma separated
                    keyword_str = match.group(1)
                    keywords.extend(k.strip() for k in re.split(r'[,\s]+', keyword_str) if k.strip())
                    
            except Exception as e:
                logger.debug(f"Error reading keywords from {file_path}: {e}")
    
    return list(set(keywords))  # Remove duplicates


def pypi_classifiers_to_tags(classifiers: List[str]) -> List[str]:
    """Convert PyPI classifiers to repoindex tags, preserving hierarchy where useful."""
    tags = []
    
    for classifier in classifiers:
        # Development Status
        if classifier.startswith('Development Status ::'):
            status_map = {
                '1 - Planning': 'planning',
                '2 - Pre-Alpha': 'pre-alpha',
                '3 - Alpha': 'alpha',
                '4 - Beta': 'beta',
                '5 - Production/Stable': 'stable',
                '6 - Mature': 'mature',
                '7 - Inactive': 'inactive'
            }
            for key, value in status_map.items():
                if key in classifier:
                    tags.append(f'status:{value}')
                    break
        
        # Programming Language - preserve version hierarchy
        elif classifier.startswith('Programming Language ::'):
            parts = classifier.split(' :: ')[1:]  # Skip "Programming Language"
            if parts and parts[0] == 'Python':
                # For Python, create hierarchical tag
                if len(parts) > 1:
                    # e.g., "Programming Language :: Python :: 3.11" -> "python:3.11"
                    tags.append(f'python:{parts[1]}')
                    # Also add generic python tag
                    tags.append('lang:python')
                else:
                    tags.append('lang:python')
            elif parts:
                # Other languages
                lang_name = parts[0].lower()
                tags.append(f'lang:{lang_name}')
                # If there's version info, add it hierarchically
                if len(parts) > 1:
                    tags.append(f'lang:{lang_name}/{parts[1].lower()}')
        
        # License
        elif classifier.startswith('License ::'):
            if 'MIT' in classifier:
                tags.append('license:mit')
            elif 'Apache' in classifier:
                if '2.0' in classifier:
                    tags.append('license:apache-2.0')
                else:
                    tags.append('license:apache')
            elif 'GPL' in classifier:
                if 'v3' in classifier or '3.0' in classifier:
                    tags.append('license:gpl-3.0')
                elif 'v2' in classifier or '2.0' in classifier:
                    tags.append('license:gpl-2.0')
                else:
                    tags.append('license:gpl')
            elif 'BSD' in classifier:
                tags.append('license:bsd')
        
        # Topic - preserve full hierarchy
        elif classifier.startswith('Topic ::'):
            parts = classifier.split(' :: ')[1:]  # Skip "Topic"
            if parts:
                # Create hierarchical tag
                # e.g., "Topic :: Scientific/Engineering :: Artificial Intelligence"
                # becomes "topic:scientific-engineering/artificial-intelligence"
                hierarchy = []
                for part in parts:
                    # Sanitize each level
                    clean_part = part.lower().replace(' ', '-').replace('/', '-')
                    hierarchy.append(clean_part)
                
                # Add the full hierarchical tag
                if len(hierarchy) > 1:
                    tags.append(f'topic:{"/".join(hierarchy)}')
                    # Also add parent levels for easier querying
                    for i in range(len(hierarchy)):
                        parent_tag = f'topic:{"/".join(hierarchy[:i+1])}'
                        if parent_tag not in tags:
                            tags.append(parent_tag)
                else:
                    tags.append(f'topic:{hierarchy[0]}')
        
        # Framework
        elif classifier.startswith('Framework ::'):
            parts = classifier.split(' :: ')[1:]
            if parts:
                framework = parts[0].lower()
                tags.append(f'framework:{framework}')
                # Add version if present
                if len(parts) > 1:
                    tags.append(f'framework:{framework}/{parts[1].lower()}')
        
        # Intended Audience
        elif classifier.startswith('Intended Audience ::'):
            audience = classifier.split(' :: ')[1].lower().replace(' ', '-').replace('/', '-')
            tags.append(f'audience:{audience}')
        
        # Operating System
        elif classifier.startswith('Operating System ::'):
            if 'OS Independent' in classifier:
                tags.append('os:cross-platform')
            elif 'POSIX' in classifier:
                tags.append('os:posix')
                if 'Linux' in classifier:
                    tags.append('os:posix/linux')
            elif 'Microsoft' in classifier:
                tags.append('os:windows')
                if 'Windows' in classifier:
                    # Extract Windows version if present
                    parts = classifier.split(' :: ')
                    for part in parts:
                        if part.startswith('Windows'):
                            tags.append(f'os:windows/{part.lower().replace(" ", "-")}')
            elif 'MacOS' in classifier:
                tags.append('os:macos')
        
        # Natural Language
        elif classifier.startswith('Natural Language ::'):
            lang = classifier.split(' :: ')[1].lower()
            tags.append(f'natural-language:{lang}')
        
        # Environment (e.g., Web Environment :: Flask)
        elif classifier.startswith('Environment ::'):
            parts = classifier.split(' :: ')[1:]
            if parts:
                env = parts[0].lower().replace(' ', '-')
                tags.append(f'environment:{env}')
                if len(parts) > 1:
                    # Add sub-environment
                    tags.append(f'environment:{env}/{parts[1].lower()}')
    
    return list(set(tags))  # Remove duplicates


def extract_pypi_tags(repo_path: str) -> List[str]:
    """Extract all PyPI-related tags from a repository."""
    tags = []
    packaging_files = find_packaging_files(repo_path)
    
    if not packaging_files:
        return tags
    
    # Extract classifiers
    classifiers = []
    for file_path in packaging_files:
        if Path(file_path).name == 'pyproject.toml':
            classifiers.extend(extract_classifiers_from_pyproject(file_path))
        elif Path(file_path).name == 'setup.py':
            classifiers.extend(extract_classifiers_from_setup_py(file_path))
        elif Path(file_path).name == 'setup.cfg':
            classifiers.extend(extract_classifiers_from_setup_cfg(file_path))
    
    # Convert classifiers to tags
    if classifiers:
        tags.extend(pypi_classifiers_to_tags(classifiers))
    
    # Add keywords as tags
    keywords = extract_keywords_from_packaging_files(repo_path)
    for keyword in keywords:
        # Sanitize keyword for use as tag
        clean_keyword = keyword.lower().replace(' ', '-').replace('_', '-')
        if clean_keyword and len(clean_keyword) <= 50:  # Reasonable length limit
            tags.append(f'keyword:{clean_keyword}')
    
    # Add package type tag
    tags.append('type:python-package')
    
    return list(set(tags))  # Remove duplicates


def update_pypi_classifier(repo_path: str, classifier_prefix: str, new_classifier: str) -> bool:
    """
    Update a PyPI classifier in packaging files.
    
    Args:
        repo_path: Path to repository
        classifier_prefix: Prefix to match (e.g., "Development Status")
        new_classifier: New classifier to set
        
    Returns:
        True if successfully updated
    """
    packaging_files = find_packaging_files(repo_path)
    updated = False
    
    for file_path in packaging_files:
        if Path(file_path).name == 'pyproject.toml':
            try:
                with open(file_path, 'rb') as f:
                    data = tomllib.load(f)
                
                # Get existing classifiers
                classifiers = []
                if 'project' in data and 'classifiers' in data['project']:
                    classifiers = data['project']['classifiers']
                elif 'tool' in data and 'setuptools' in data['tool'] and 'classifiers' in data['tool']['setuptools']:
                    classifiers = data['tool']['setuptools']['classifiers']
                else:
                    # No classifiers section, create one
                    if 'project' not in data:
                        data['project'] = {}
                    data['project']['classifiers'] = []
                    classifiers = data['project']['classifiers']
                
                # Remove existing classifiers with the same prefix
                classifiers = [c for c in classifiers if not c.startswith(classifier_prefix)]
                
                # Add new classifier
                classifiers.append(new_classifier)
                classifiers.sort()  # Keep them sorted
                
                # Update the data
                if 'project' in data and 'classifiers' in data['project']:
                    data['project']['classifiers'] = classifiers
                elif 'tool' in data and 'setuptools' in data['tool']:
                    data['tool']['setuptools']['classifiers'] = classifiers
                
                # Write back
                import toml
                with open(file_path, 'w') as f:
                    toml.dump(data, f)
                
                updated = True
                logger.info(f"Updated classifier in {file_path}")
                
            except Exception as e:
                logger.error(f"Error updating classifier in {file_path}: {e}")
    
    return updated


def update_pypi_license(repo_path: str, license_key: str) -> bool:
    """
    Update license in packaging files.
    
    Args:
        repo_path: Path to repository
        license_key: License identifier (e.g., "mit", "apache-2.0")
        
    Returns:
        True if successfully updated
    """
    # Map common license keys to PyPI classifiers
    license_map = {
        'mit': 'License :: OSI Approved :: MIT License',
        'apache': 'License :: OSI Approved :: Apache Software License',
        'apache-2.0': 'License :: OSI Approved :: Apache Software License',
        'gpl': 'License :: OSI Approved :: GNU General Public License (GPL)',
        'gpl-2.0': 'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'gpl-3.0': 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'bsd': 'License :: OSI Approved :: BSD License',
        'bsd-3-clause': 'License :: OSI Approved :: BSD License',
        'mpl-2.0': 'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'lgpl': 'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    }
    
    if license_key in license_map:
        return update_pypi_classifier(repo_path, "License ::", license_map[license_key])
    
    return False


def sync_tag_to_pypi(repo_path: str, tag: str) -> bool:
    """
    Sync a repoindex tag to PyPI metadata.
    
    Args:
        repo_path: Path to repository
        tag: Tag to sync (e.g., "status:beta", "license:mit")
        
    Returns:
        True if successfully synced
    """
    # Parse tag
    if ':' not in tag:
        return False
    
    key, value = tag.split(':', 1)
    
    if key == 'status':
        # Map status to PyPI development status
        status_map = {
            'planning': 'Development Status :: 1 - Planning',
            'pre-alpha': 'Development Status :: 2 - Pre-Alpha',
            'alpha': 'Development Status :: 3 - Alpha',
            'beta': 'Development Status :: 4 - Beta',
            'stable': 'Development Status :: 5 - Production/Stable',
            'mature': 'Development Status :: 6 - Mature',
            'inactive': 'Development Status :: 7 - Inactive'
        }
        
        if value in status_map:
            return update_pypi_classifier(repo_path, "Development Status ::", status_map[value])
    
    elif key == 'license':
        return update_pypi_license(repo_path, value)
    
    elif key == 'python':
        # Update Python version classifier
        classifier = f"Programming Language :: Python :: {value}"
        return update_pypi_classifier(repo_path, f"Programming Language :: Python :: {value}", classifier)
    
    elif key == 'framework':
        # Handle framework tags
        if '/' in value:
            # Hierarchical framework tag
            parts = value.split('/')
            framework = parts[0].title()
            version = parts[1] if len(parts) > 1 else None
            
            if version:
                classifier = f"Framework :: {framework} :: {version}"
            else:
                classifier = f"Framework :: {framework}"
        else:
            classifier = f"Framework :: {value.title()}"
        
        return update_pypi_classifier(repo_path, f"Framework :: {value.title()}", classifier)
    
    elif key == 'audience':
        # Map audience tags
        audience_map = {
            'developers': 'Intended Audience :: Developers',
            'science-research': 'Intended Audience :: Science/Research',
            'system-administrators': 'Intended Audience :: System Administrators',
            'end-users': 'Intended Audience :: End Users/Desktop',
            'education': 'Intended Audience :: Education'
        }
        
        if value in audience_map:
            return update_pypi_classifier(repo_path, "Intended Audience ::", audience_map[value])
    
    return False


def sync_pypi_tags(repo_path: str, tags: List[str]) -> Dict[str, bool]:
    """
    Sync multiple tags to PyPI metadata.
    
    Args:
        repo_path: Path to repository
        tags: List of tags to sync
        
    Returns:
        Dictionary mapping tags to success status
    """
    results = {}
    
    for tag in tags:
        results[tag] = sync_tag_to_pypi(repo_path, tag)
    
    return results


def is_package_outdated(repo_path: str, package_name: str, pypi_version: str) -> bool:
    """Check if local package version is behind PyPI version."""
    local_version = get_local_package_version(repo_path, package_name)
    
    if not local_version:
        return False
    
    try:
        # Simple version comparison (you might want to use packaging.version for more robust comparison)
        from packaging import version
        return version.parse(local_version) < version.parse(pypi_version)
    except ImportError:
        # Fallback to string comparison if packaging module not available
        return local_version != pypi_version
    except Exception:
        return False


"""
Unit tests for ghops shell commands.

Tests for newly implemented shell commands:
- do_git() with -r recursive flag
- do_clone() with various options
- do_config() with subcommands
- do_export() with different formats
- do_docs() with detect/build/deploy (recently fixed imports)
"""

import pytest
import json
import tempfile
import os
from pathlib import Path
from unittest.mock import MagicMock, patch, call, ANY
import click

from repoindex.shell.shell import RepoIndexShell


@pytest.fixture
def temp_config_dir(tmp_path):
    """Create temporary config directory."""
    config_dir = tmp_path / ".repoindex"
    config_dir.mkdir()
    return config_dir


@pytest.fixture
def mock_repos(tmp_path):
    """Create mock git repositories for testing."""
    repos = []
    for i, name in enumerate(['project-a', 'project-b', 'project-c']):
        repo_path = tmp_path / name
        repo_path.mkdir()
        (repo_path / '.git').mkdir()
        repos.append(str(repo_path))
    return repos


@pytest.fixture
def mock_config(mock_repos):
    """Create mock config with test repositories."""
    return {
        'general': {
            'repository_directories': [str(Path(mock_repos[0]).parent)]
        },
        'repository_tags': {
            mock_repos[0]: ['alex/beta', 'topic:ml'],
            mock_repos[1]: ['alex/production', 'lang:python'],
            mock_repos[2]: ['topic:scientific/engineering/ai']
        }
    }


@pytest.fixture
def mock_metadata():
    """Create mock metadata store."""
    return {
        'language': 'Python',
        'status': {
            'has_uncommitted_changes': False,
            'branch': 'main'
        }
    }


@pytest.fixture
def shell_instance(mock_repos):
    """Create a shell instance with mocked dependencies."""
    with patch('repoindex.shell.shell.load_config') as mock_load_config, \
         patch('repoindex.shell.shell.find_git_repos_from_config') as mock_find_repos, \
         patch('repoindex.shell.shell.get_metadata_store') as mock_metadata_store, \
         patch('repoindex.shell.shell.get_repository_tags') as mock_get_tags:

        mock_load_config.return_value = {
            'general': {'repository_directories': [str(Path(mock_repos[0]).parent)]},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        shell = RepoIndexShell()
        return shell


class TestShellGitRecursive:
    """Test git command with recursive flag."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.git_ops.utils.get_repos_from_vfs_path')
    @patch('repoindex.commands.git.git_status')
    def test_git_recursive_status_multiple_repos(self, mock_git_status, mock_get_repos,
                                                  mock_get_tags, mock_metadata_store,
                                                  mock_find_repos, mock_load_config,
                                                  mock_repos, capsys):
        """Test git -r status on multiple repos."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock VFS resolution to return all repos
        mock_get_repos.return_value = mock_repos

        # Mock git_status callback
        mock_git_status.callback = MagicMock()

        # Create shell and execute
        shell = RepoIndexShell()
        shell.do_git('-r status')

        # Verify get_repos_from_vfs_path was called
        mock_get_repos.assert_called_once_with('/')

        # Verify git_status was called for each repo
        assert mock_git_status.callback.call_count == len(mock_repos)

        # Check output shows progress
        captured = capsys.readouterr()
        assert f"Running 'git status' on {len(mock_repos)} repositories" in captured.out
        assert 'project-a' in captured.out or 'project-b' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.git_ops.utils.get_repos_from_vfs_path')
    def test_git_recursive_pull_on_vfs_path(self, mock_get_repos,
                                             mock_get_tags, mock_metadata_store,
                                             mock_find_repos, mock_load_config,
                                             mock_repos, capsys):
        """Test git -r pull shows deprecation message (pull removed)."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock VFS resolution
        mock_get_repos.return_value = mock_repos

        # Create shell
        shell = RepoIndexShell()

        # Change to specific VFS path
        shell.do_cd('/repos')

        # Execute git pull - should show error (command removed)
        shell.do_git('pull')

        # Verify error message shown
        captured = capsys.readouterr()
        assert 'not a supported git command' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.git.git_status')
    def test_git_status_non_recursive(self, mock_git_status, mock_get_tags,
                                       mock_metadata_store, mock_find_repos,
                                       mock_load_config, mock_repos):
        """Test git status (non-recursive) on single repo."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock git_status callback
        mock_git_status.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute non-recursive status
        shell.do_git('status')

        # Should be called once for current path
        mock_git_status.callback.assert_called_once()

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.git_ops.utils.get_repos_from_vfs_path')
    def test_git_recursive_no_repos(self, mock_get_repos, mock_get_tags,
                                     mock_metadata_store, mock_find_repos,
                                     mock_load_config, capsys):
        """Test git -r when VFS path has no repos."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock VFS resolution returning no repos
        mock_get_repos.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute recursive git command
        shell.do_git('-r status')

        # Check error message
        captured = capsys.readouterr()
        assert 'No repositories found' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.git.git_log')
    def test_git_log_with_args(self, mock_git_log, mock_get_tags,
                               mock_metadata_store, mock_find_repos,
                               mock_load_config, mock_repos):
        """Test git log with additional arguments."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock git_log callback
        mock_git_log.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute git log with flags
        shell.do_git('log --oneline -n 5')

        # Verify callback was called with correct args
        mock_git_log.callback.assert_called_once()
        call_args = mock_git_log.callback.call_args
        # Shell calls callback with positional args:
        # git_log.callback(vfs_path, oneline, max_count, since, author, graph, all_branches, False)
        assert len(call_args[0]) >= 3  # At least vfs_path, oneline, max_count
        # Second arg is oneline (should be True)
        assert call_args[0][1] == True  # oneline
        # Third arg is max_count (should be 5)
        assert call_args[0][2] == 5  # max_count


class TestShellClone:
    """Test clone command."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_single_repo(self, mock_handler, mock_get_tags,
                               mock_metadata_store, mock_find_repos,
                               mock_load_config, capsys):
        """Test cloning single repo."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler
        mock_handler.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute clone
        shell.do_clone('user/repo')

        # Verify clone_handler was called
        mock_handler.callback.assert_called_once()
        call_args = mock_handler.callback.call_args
        assert call_args[1]['repos'] == ['user/repo']
        assert call_args[1]['user'] is None

        # Check VFS refresh message
        captured = capsys.readouterr()
        assert 'Refreshing VFS' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_multiple_repos(self, mock_handler, mock_get_tags,
                                   mock_metadata_store, mock_find_repos,
                                   mock_load_config):
        """Test cloning multiple repos."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler
        mock_handler.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute clone with multiple repos
        shell.do_clone('user/repo1 user/repo2 user/repo3')

        # Verify clone_handler was called with all repos
        mock_handler.callback.assert_called_once()
        call_args = mock_handler.callback.call_args
        assert call_args[1]['repos'] == ['user/repo1', 'user/repo2', 'user/repo3']

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_user_repos(self, mock_handler, mock_get_tags,
                              mock_metadata_store, mock_find_repos,
                              mock_load_config):
        """Test cloning all repos for user."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler
        mock_handler.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute clone --user
        shell.do_clone('--user username')

        # Verify clone_handler was called with user
        mock_handler.callback.assert_called_once()
        call_args = mock_handler.callback.call_args
        assert call_args[1]['user'] == 'username'
        assert call_args[1]['repos'] == []
        assert call_args[1]['limit'] == 100

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_user_with_limit(self, mock_handler, mock_get_tags,
                                    mock_metadata_store, mock_find_repos,
                                    mock_load_config):
        """Test cloning user repos with limit."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler
        mock_handler.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute clone --user with --limit
        shell.do_clone('--user username --limit 10')

        # Verify clone_handler was called with correct limit
        mock_handler.callback.assert_called_once()
        call_args = mock_handler.callback.call_args
        assert call_args[1]['user'] == 'username'
        assert call_args[1]['limit'] == 10

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_url(self, mock_handler, mock_get_tags,
                       mock_metadata_store, mock_find_repos,
                       mock_load_config):
        """Test cloning from URL."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler
        mock_handler.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute clone with URL
        shell.do_clone('https://github.com/user/repo')

        # Verify clone_handler was called with URL
        mock_handler.callback.assert_called_once()
        call_args = mock_handler.callback.call_args
        assert call_args[1]['repos'] == ['https://github.com/user/repo']

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_clone_no_args(self, mock_get_tags, mock_metadata_store,
                          mock_find_repos, mock_load_config, capsys):
        """Test clone with no arguments."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute clone without args
        shell.do_clone('')

        # Check error message
        captured = capsys.readouterr()
        assert 'Usage:' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_error_handling(self, mock_handler, mock_get_tags,
                                   mock_metadata_store, mock_find_repos,
                                   mock_load_config, capsys):
        """Test clone error handling."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock clone handler to raise exception
        mock_handler.callback = MagicMock(side_effect=Exception("Clone failed"))

        # Create shell
        shell = RepoIndexShell()

        # Execute clone
        shell.do_clone('user/repo')

        # Check error message
        captured = capsys.readouterr()
        assert 'error' in captured.out.lower()


class TestShellConfig:
    """Test config command."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config.show_config')
    def test_config_show(self, mock_show_config, mock_get_tags,
                        mock_metadata_store, mock_find_repos,
                        mock_load_config):
        """Test config show command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock show_config
        mock_show_config.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute config show
        shell.do_config('show')

        # Verify show_config was called
        mock_show_config.callback.assert_called_once()

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config_repos.repos_list')
    def test_config_repos_list(self, mock_repos_list, mock_get_tags,
                               mock_metadata_store, mock_find_repos,
                               mock_load_config):
        """Test config repos list command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock repos_list
        mock_repos_list.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute config repos list
        shell.do_config('repos list')

        # Verify repos_list was called
        mock_repos_list.callback.assert_called_once()
        call_args = mock_repos_list.callback.call_args
        assert call_args[1]['json_output'] == False

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config_repos.repos_add')
    def test_config_repos_add(self, mock_repos_add, mock_get_tags,
                             mock_metadata_store, mock_find_repos,
                             mock_load_config, capsys):
        """Test config repos add command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock repos_add
        mock_repos_add.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute config repos add
        shell.do_config('repos add /path/to/repos')

        # Verify repos_add was called
        mock_repos_add.callback.assert_called_once()
        call_args = mock_repos_add.callback.call_args
        assert call_args[1]['path'] == '/path/to/repos'
        assert call_args[1]['refresh'] == False

        # Check VFS refresh message
        captured = capsys.readouterr()
        assert 'Refreshing VFS' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config_repos.repos_remove')
    def test_config_repos_remove(self, mock_repos_remove, mock_get_tags,
                                 mock_metadata_store, mock_find_repos,
                                 mock_load_config, capsys):
        """Test config repos remove command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock repos_remove
        mock_repos_remove.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute config repos remove
        shell.do_config('repos remove /path/to/repos')

        # Verify repos_remove was called
        mock_repos_remove.callback.assert_called_once()
        call_args = mock_repos_remove.callback.call_args
        assert call_args[1]['path'] == '/path/to/repos'

        # Check VFS refresh message
        captured = capsys.readouterr()
        assert 'Refreshing VFS' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config_repos.repos_clear')
    def test_config_repos_clear(self, mock_repos_clear, mock_get_tags,
                                mock_metadata_store, mock_find_repos,
                                mock_load_config, capsys):
        """Test config repos clear command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock repos_clear
        mock_repos_clear.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Execute config repos clear
        shell.do_config('repos clear')

        # Verify repos_clear was called
        mock_repos_clear.callback.assert_called_once()

        # Check VFS refresh message
        captured = capsys.readouterr()
        assert 'Refreshing VFS' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_config_no_args(self, mock_get_tags, mock_metadata_store,
                           mock_find_repos, mock_load_config, capsys):
        """Test config with no arguments."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute config without args
        shell.do_config('')

        # Check error message
        captured = capsys.readouterr()
        assert 'Usage:' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_config_invalid_subcommand(self, mock_get_tags, mock_metadata_store,
                                       mock_find_repos, mock_load_config, capsys):
        """Test config with invalid subcommand."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute config with invalid subcommand
        shell.do_config('invalid')

        # Check error message
        captured = capsys.readouterr()
        assert 'Unknown' in captured.out


class TestShellExport:
    """Test export command (now shows help about using VFS for exports)."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_export_shows_vfs_help(self, mock_get_tags, mock_metadata_store,
                                   mock_find_repos, mock_load_config, capsys):
        """Test export command shows VFS-based export guidance."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute export with any args - should show guidance
        shell.do_export('markdown')

        # Check output contains VFS guidance
        captured = capsys.readouterr()
        assert 'VFS' in captured.out
        assert 'metadata' in captured.out.lower() or 'JSON' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_export_no_args(self, mock_get_tags, mock_metadata_store,
                           mock_find_repos, mock_load_config, capsys):
        """Test export with no arguments shows guidance."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute export without args
        shell.do_export('')

        # Check output contains guidance
        captured = capsys.readouterr()
        assert 'VFS' in captured.out or 'external tools' in captured.out


class TestShellDocs:
    """Test docs command (simplified detection-only)."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.docs.get_docs_status')
    def test_docs_detection(self, mock_get_docs_status, mock_get_tags,
                           mock_metadata_store, mock_find_repos,
                           mock_load_config, mock_repos, capsys):
        """Test docs detection command."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock get_docs_status
        mock_get_docs_status.return_value = {
            'name': 'project-a',
            'has_docs': True,
            'docs_tool': 'mkdocs',
            'docs_config': 'mkdocs.yml'
        }

        # Create shell
        shell = RepoIndexShell()

        # Navigate to repo
        shell.do_cd('/repos/project-a')

        # Execute docs
        shell.do_docs('')

        # Verify get_docs_status was called
        mock_get_docs_status.assert_called()

        # Check output
        captured = capsys.readouterr()
        assert 'mkdocs' in captured.out or 'project-a' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.docs.get_docs_status')
    def test_docs_no_docs(self, mock_get_docs_status, mock_get_tags,
                         mock_metadata_store, mock_find_repos,
                         mock_load_config, mock_repos, capsys):
        """Test docs when no documentation found."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock get_docs_status - no docs
        mock_get_docs_status.return_value = {
            'name': 'project-a',
            'has_docs': False,
            'docs_tool': None,
            'docs_config': None
        }

        # Create shell
        shell = RepoIndexShell()

        # Navigate to repo
        shell.do_cd('/repos/project-a')

        # Execute docs
        shell.do_docs('')

        # Check output
        captured = capsys.readouterr()
        assert 'no documentation detected' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.git_ops.utils.get_repos_from_vfs_path')
    def test_docs_no_repo_found(self, mock_get_repos_from_vfs, mock_get_tags,
                                mock_metadata_store, mock_find_repos,
                                mock_load_config, capsys):
        """Test docs when no repository found in current directory."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock VFS path resolution returning no repos
        mock_get_repos_from_vfs.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Stay at root (no repo)
        shell.do_docs('')

        # Check error message
        captured = capsys.readouterr()
        assert 'No repository' in captured.out


class TestShellEdgeCases:
    """Test edge cases and error handling."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_git_no_args(self, mock_get_tags, mock_metadata_store,
                        mock_find_repos, mock_load_config, capsys):
        """Test git with no arguments."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute git without args
        shell.do_git('')

        # Check error message
        captured = capsys.readouterr()
        assert 'Usage:' in captured.out
        assert 'Supported:' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    def test_git_unsupported_command(self, mock_get_tags, mock_metadata_store,
                                     mock_find_repos, mock_load_config,
                                     mock_repos, capsys):
        """Test git with unsupported subcommand."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute git with unsupported command
        shell.do_git('rebase')

        # Check error message
        captured = capsys.readouterr()
        assert 'not a supported' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.clone.clone_handler')
    def test_clone_user_missing_username(self, mock_handler, mock_get_tags,
                                         mock_metadata_store, mock_find_repos,
                                         mock_load_config, capsys):
        """Test clone --user without username."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute clone --user without username
        shell.do_clone('--user')

        # Check error message
        captured = capsys.readouterr()
        assert 'Error' in captured.out or 'requires a username' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.config_repos.repos_add')
    def test_config_repos_add_no_path(self, mock_repos_add, mock_get_tags,
                                      mock_metadata_store, mock_find_repos,
                                      mock_load_config, capsys):
        """Test config repos add without path."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = []

        mock_store = MagicMock()
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Create shell
        shell = RepoIndexShell()

        # Execute config repos add without path
        shell.do_config('repos add')

        # Check error message
        captured = capsys.readouterr()
        assert 'Usage:' in captured.out


class TestShellIntegration:
    """Integration tests for shell commands."""

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.git_ops.utils.get_repos_from_vfs_path')
    @patch('repoindex.commands.git.git_status')
    def test_git_recursive_from_different_vfs_paths(self, mock_git_status,
                                                     mock_get_repos, mock_get_tags,
                                                     mock_metadata_store,
                                                     mock_find_repos,
                                                     mock_load_config,
                                                     mock_repos, capsys):
        """Test git -r status from different VFS paths."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock git_status
        mock_git_status.callback = MagicMock()

        # Create shell
        shell = RepoIndexShell()

        # Test from /by-language/Python
        shell.do_cd('/by-language')
        mock_get_repos.return_value = [mock_repos[0], mock_repos[1]]  # Two Python repos
        shell.do_git('-r status')

        # Should operate on repos from that VFS path
        captured = capsys.readouterr()
        assert 'Running' in captured.out
        assert 'repositories' in captured.out

    @patch('repoindex.shell.shell.load_config')
    @patch('repoindex.shell.shell.find_git_repos_from_config')
    @patch('repoindex.shell.shell.get_metadata_store')
    @patch('repoindex.shell.shell.get_repository_tags')
    @patch('repoindex.commands.docs.get_docs_status')
    def test_docs_in_real_fs_mode(self, mock_get_docs_status, mock_get_tags,
                                  mock_metadata_store, mock_find_repos,
                                  mock_load_config, mock_repos, capsys):
        """Test docs command when in real filesystem mode."""
        # Setup mocks
        mock_load_config.return_value = {
            'general': {'repository_directories': ['/tmp/repos']},
            'repository_tags': {}
        }
        mock_find_repos.return_value = mock_repos

        mock_store = MagicMock()
        mock_store.get.return_value = {'language': 'Python', 'status': {}}
        mock_metadata_store.return_value = mock_store

        mock_get_tags.return_value = []

        # Mock get_docs_status
        mock_get_docs_status.return_value = {
            'name': 'project-a',
            'has_docs': True,
            'docs_tool': 'sphinx',
            'docs_config': 'docs/conf.py'
        }

        # Create shell
        shell = RepoIndexShell()

        # Navigate into repo (enters real filesystem mode)
        shell.do_cd('/repos/project-a')

        # Set real filesystem state manually for testing
        shell.in_real_fs = True
        shell.real_fs_repo = mock_repos[0]

        # Execute docs command
        shell.do_docs('')

        # Should use real_fs_repo path
        mock_get_docs_status.assert_called_once()

        # Check output
        captured = capsys.readouterr()
        assert 'sphinx' in captured.out or 'project-a' in captured.out



"""
Activity monitoring and event tracking for TUI.
"""

import asyncio
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, List, Dict, Any, Set
from enum import Enum
import json

# Debug log file
DEBUG_LOG = open("/tmp/repoindex_watcher_debug.log", "a")

def debug_log(msg: str):
    """Write to debug log."""
    DEBUG_LOG.write(f"{msg}\n")
    DEBUG_LOG.flush()


class EventType(Enum):
    """Types of repository events."""
    FILE_MODIFIED = "modified"
    FILE_CREATED = "created"
    FILE_DELETED = "deleted"
    GIT_COMMIT = "commit"
    GIT_PUSH = "push"
    GIT_PULL = "pull"
    TEST_STARTED = "test_started"
    TEST_PASSED = "test_passed"
    TEST_FAILED = "test_failed"
    BUILD_STARTED = "build_started"
    BUILD_SUCCESS = "build_success"
    BUILD_FAILED = "build_failed"


@dataclass
class ActivityEvent:
    """Represents a single activity event."""
    timestamp: datetime
    repo_path: str
    repo_name: str
    event_type: EventType
    file_path: Optional[str] = None
    message: str = ""
    details: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        """Ensure timestamp is timezone-aware."""
        if self.timestamp.tzinfo is None:
            self.timestamp = self.timestamp.replace(tzinfo=timezone.utc)

    @property
    def age_seconds(self) -> float:
        """Get age of event in seconds."""
        now = datetime.now(timezone.utc)
        return (now - self.timestamp).total_seconds()

    @property
    def age_display(self) -> str:
        """Get human-readable age."""
        seconds = self.age_seconds

        if seconds < 60:
            return f"{int(seconds)}s ago"
        elif seconds < 3600:
            return f"{int(seconds / 60)}m ago"
        elif seconds < 86400:
            return f"{int(seconds / 3600)}h ago"
        else:
            return f"{int(seconds / 86400)}d ago"

    @property
    def icon(self) -> str:
        """Get icon for event type."""
        icons = {
            EventType.FILE_MODIFIED: "🟢",
            EventType.FILE_CREATED: "🟡",
            EventType.FILE_DELETED: "🔴",
            EventType.GIT_COMMIT: "🔵",
            EventType.GIT_PUSH: "📤",
            EventType.GIT_PULL: "📥",
            EventType.TEST_STARTED: "🧪",
            EventType.TEST_PASSED: "✅",
            EventType.TEST_FAILED: "❌",
            EventType.BUILD_STARTED: "🔨",
            EventType.BUILD_SUCCESS: "✅",
            EventType.BUILD_FAILED: "❌",
        }
        return icons.get(self.event_type, "⚪")

    @property
    def color(self) -> str:
        """Get color for event type."""
        colors = {
            EventType.FILE_MODIFIED: "green",
            EventType.FILE_CREATED: "yellow",
            EventType.FILE_DELETED: "red",
            EventType.GIT_COMMIT: "blue",
            EventType.GIT_PUSH: "cyan",
            EventType.GIT_PULL: "cyan",
            EventType.TEST_STARTED: "yellow",
            EventType.TEST_PASSED: "green",
            EventType.TEST_FAILED: "red",
            EventType.BUILD_STARTED: "yellow",
            EventType.BUILD_SUCCESS: "green",
            EventType.BUILD_FAILED: "red",
        }
        return colors.get(self.event_type, "white")

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "timestamp": self.timestamp.isoformat(),
            "repo_path": self.repo_path,
            "repo_name": self.repo_name,
            "event_type": self.event_type.value,
            "file_path": self.file_path,
            "message": self.message,
            "details": self.details,
        }


class ActivityFeed:
    """Manages the activity feed and event history."""

    def __init__(self, max_events: int = 500):
        """Initialize activity feed.

        Args:
            max_events: Maximum number of events to keep in memory
        """
        self.max_events = max_events
        self.events: List[ActivityEvent] = []
        self.listeners: List[callable] = []
        self._lock = asyncio.Lock()

    async def add_event(self, event: ActivityEvent):
        """Add event to feed.

        Args:
            event: Event to add
        """
        debug_log(f"[FEED] add_event called: {event.repo_name}/{event.file_path or event.message}")
        async with self._lock:
            self.events.insert(0, event)  # Most recent first
            debug_log(f"[FEED] Event added to feed, total events: {len(self.events)}")

            # Trim old events
            if len(self.events) > self.max_events:
                self.events = self.events[:self.max_events]

            # Notify listeners
            debug_log(f"[FEED] Notifying {len(self.listeners)} listeners")
            for i, listener in enumerate(self.listeners):
                try:
                    debug_log(f"[FEED] Calling listener {i}")
                    if asyncio.iscoroutinefunction(listener):
                        await listener(event)
                    else:
                        listener(event)
                    debug_log(f"[FEED] Listener {i} completed")
                except Exception as e:
                    # Don't let listener errors break the feed
                    debug_log(f"[FEED] Listener {i} error: {e}")

    def add_listener(self, callback: callable):
        """Add event listener.

        Args:
            callback: Function to call when event is added
        """
        self.listeners.append(callback)

    def remove_listener(self, callback: callable):
        """Remove event listener.

        Args:
            callback: Listener to remove
        """
        if callback in self.listeners:
            self.listeners.remove(callback)

    async def get_events(
        self,
        limit: Optional[int] = None,
        repo_path: Optional[str] = None,
        event_types: Optional[Set[EventType]] = None
    ) -> List[ActivityEvent]:
        """Get events with optional filtering.

        Args:
            limit: Maximum number of events to return
            repo_path: Filter by repository path
            event_types: Filter by event types

        Returns:
            List of events
        """
        async with self._lock:
            events = self.events

            # Filter by repo
            if repo_path:
                events = [e for e in events if e.repo_path == repo_path]

            # Filter by event type
            if event_types:
                events = [e for e in events if e.event_type in event_types]

            # Limit results
            if limit:
                events = events[:limit]

            return events

    async def clear(self):
        """Clear all events."""
        async with self._lock:
            self.events.clear()

    async def export_json(self, file_path: str):
        """Export events to JSON file.

        Args:
            file_path: Path to output file
        """
        async with self._lock:
            data = [event.to_dict() for event in self.events]
            with open(file_path, 'w') as f:
                json.dump(data, f, indent=2)


class RepositoryStats:
    """Track repository statistics for the dashboard."""

    def __init__(self):
        """Initialize stats tracker."""
        self.dirty_repos: Set[str] = set()
        self.unpushed_repos: Set[str] = set()
        self.active_repos: Set[str] = set()  # Active in last 5 minutes
        self.failing_tests: Set[str] = set()
        self.repo_last_activity: Dict[str, datetime] = {}  # Track last activity time
        self._lock = asyncio.Lock()

    async def mark_dirty(self, repo_path: str, activity_time: Optional[datetime] = None):
        """Mark repository as having uncommitted changes."""
        async with self._lock:
            self.dirty_repos.add(repo_path)
            if activity_time:
                self.repo_last_activity[repo_path] = activity_time

    async def mark_clean(self, repo_path: str):
        """Mark repository as clean."""
        async with self._lock:
            self.dirty_repos.discard(repo_path)

    async def mark_unpushed(self, repo_path: str, activity_time: Optional[datetime] = None):
        """Mark repository as having unpushed commits."""
        async with self._lock:
            self.unpushed_repos.add(repo_path)
            if activity_time:
                self.repo_last_activity[repo_path] = activity_time

    async def mark_pushed(self, repo_path: str):
        """Mark repository as pushed."""
        async with self._lock:
            self.unpushed_repos.discard(repo_path)

    async def mark_active(self, repo_path: str):
        """Mark repository as recently active."""
        async with self._lock:
            self.active_repos.add(repo_path)

    async def mark_test_failed(self, repo_path: str):
        """Mark repository as having failing tests."""
        async with self._lock:
            self.failing_tests.add(repo_path)

    async def mark_test_passed(self, repo_path: str):
        """Mark repository tests as passing."""
        async with self._lock:
            self.failing_tests.discard(repo_path)

    async def get_stats(self) -> Dict[str, Any]:
        """Get current statistics.

        Returns:
            Dictionary of stats
        """
        async with self._lock:
            return {
                "dirty": len(self.dirty_repos),
                "unpushed": len(self.unpushed_repos),
                "active": len(self.active_repos),
                "failing_tests": len(self.failing_tests),
            }

    async def get_attention_repos(self) -> List[Dict[str, Any]]:
        """Get repositories that need attention.

        Returns:
            List of repo info dicts
        """
        async with self._lock:
            attention = []

            # Add dirty repos
            for repo in self.dirty_repos:
                attention.append({
                    "path": repo,
                    "name": Path(repo).name,
                    "issues": ["uncommitted changes"],
                    "priority": 2,
                    "last_activity": self.repo_last_activity.get(repo, datetime.min.replace(tzinfo=timezone.utc))
                })

            # Add unpushed repos
            for repo in self.unpushed_repos:
                existing = next((a for a in attention if a["path"] == repo), None)
                if existing:
                    existing["issues"].append("unpushed commits")
                    existing["priority"] = max(existing["priority"], 2)
                else:
                    attention.append({
                        "path": repo,
                        "name": Path(repo).name,
                        "issues": ["unpushed commits"],
                        "priority": 2,
                        "last_activity": self.repo_last_activity.get(repo, datetime.min.replace(tzinfo=timezone.utc))
                    })

            # Add failing tests (highest priority)
            for repo in self.failing_tests:
                existing = next((a for a in attention if a["path"] == repo), None)
                if existing:
                    existing["issues"].append("tests failing")
                    existing["priority"] = 1  # Highest priority
                else:
                    attention.append({
                        "path": repo,
                        "name": Path(repo).name,
                        "issues": ["tests failing"],
                        "priority": 1,
                        "last_activity": self.repo_last_activity.get(repo, datetime.min.replace(tzinfo=timezone.utc))
                    })

            # Sort by priority (highest first), then by most recent activity
            attention.sort(key=lambda x: (x["priority"], -x["last_activity"].timestamp()))

            return attention



"""
File system watcher for repository activity monitoring.
"""

import asyncio
from pathlib import Path
from typing import Dict, Set, Optional, List
from datetime import datetime, timezone
import subprocess
import sys

# Debug log file
DEBUG_LOG = open("/tmp/repoindex_watcher_debug.log", "a")

def debug_log(msg: str):
    """Write to debug log."""
    DEBUG_LOG.write(f"{msg}\n")
    DEBUG_LOG.flush()

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler, FileSystemEvent
    WATCHDOG_AVAILABLE = True
except ImportError:
    WATCHDOG_AVAILABLE = False
    Observer = None
    FileSystemEventHandler = None
    FileSystemEvent = None

from .activity import ActivityFeed, ActivityEvent, EventType, RepositoryStats


class RepositoryEventHandler(FileSystemEventHandler):
    """Handles file system events for a single repository."""

    def __init__(self, repo_path: str, repo_name: str, activity_feed: ActivityFeed, stats: RepositoryStats, loop: asyncio.AbstractEventLoop):
        """Initialize event handler.

        Args:
            repo_path: Path to repository
            repo_name: Repository name
            activity_feed: Activity feed to add events to
            stats: Stats tracker
            loop: Event loop to use for async operations
        """
        super().__init__()
        self.repo_path = repo_path
        self.repo_name = repo_name
        self.activity_feed = activity_feed
        self.stats = stats
        self.loop = loop

        # Track git status
        self._last_git_check = None
        self._is_dirty = False

    def _should_ignore(self, path: str) -> bool:
        """Check if path should be ignored.

        Args:
            path: File path to check

        Returns:
            True if should be ignored
        """
        ignore_patterns = [
            '/.git/',
            '/__pycache__/',
            '/node_modules/',
            '/.venv/',
            '/venv/',
            '/build/',
            '/dist/',
            '/.pytest_cache/',
            '/.mypy_cache/',
            '/.tox/',
            '.pyc',
            '.pyo',
            '.swp',
            '.tmp',
            '~',
        ]

        for pattern in ignore_patterns:
            if pattern in path or path.endswith(pattern.lstrip('/')):
                return True

        return False

    def _get_relative_path(self, full_path: str) -> str:
        """Get path relative to repository root.

        Args:
            full_path: Full file path

        Returns:
            Relative path
        """
        try:
            return str(Path(full_path).relative_to(self.repo_path))
        except ValueError:
            return full_path

    def _add_event(self, event_type: EventType, file_path: Optional[str] = None, message: str = ""):
        """Add event to activity feed.

        Args:
            event_type: Type of event
            file_path: Optional file path
            message: Event message
        """
        debug_log(f"[WATCHER] _add_event called: {event_type} {self.repo_name}/{file_path or message}")

        event = ActivityEvent(
            timestamp=datetime.now(timezone.utc),
            repo_path=self.repo_path,
            repo_name=self.repo_name,
            event_type=event_type,
            file_path=file_path,
            message=message
        )

        debug_log(f"[WATCHER] Event created, adding to feed (loop={self.loop})")

        # Add to feed asynchronously
        try:
            future = asyncio.run_coroutine_threadsafe(
                self.activity_feed.add_event(event),
                self.loop
            )
            debug_log(f"[WATCHER] Scheduled add_event, future={future}")
            # Don't wait for result to avoid blocking the watchdog thread
        except Exception as e:
            debug_log(f"[WATCHER] ERROR scheduling add_event: {e}")

        # Update stats
        if event_type in [EventType.FILE_MODIFIED, EventType.FILE_CREATED, EventType.FILE_DELETED]:
            try:
                asyncio.run_coroutine_threadsafe(
                    self.stats.mark_dirty(self.repo_path),
                    self.loop
                )
                asyncio.run_coroutine_threadsafe(
                    self.stats.mark_active(self.repo_path),
                    self.loop
                )
            except Exception as e:
                debug_log(f"[WATCHER] ERROR scheduling stats update: {e}")

    def on_modified(self, event: FileSystemEvent):
        """Handle file modification."""
        # Debug logging
        debug_log(f"[WATCHER] on_modified: {event.src_path}")

        if event.is_directory:
            debug_log(f"[WATCHER] Ignoring directory: {event.src_path}")
            return

        if self._should_ignore(event.src_path):
            debug_log(f"[WATCHER] Ignoring pattern match: {event.src_path}")
            return

        relative_path = self._get_relative_path(event.src_path)
        debug_log(f"[WATCHER] Adding FILE_MODIFIED event: {self.repo_name}/{relative_path}")
        self._add_event(
            EventType.FILE_MODIFIED,
            relative_path,
            f"{relative_path} modified"
        )

    def on_created(self, event: FileSystemEvent):
        """Handle file creation."""
        if event.is_directory or self._should_ignore(event.src_path):
            return

        relative_path = self._get_relative_path(event.src_path)
        self._add_event(
            EventType.FILE_CREATED,
            relative_path,
            f"{relative_path} created"
        )

    def on_deleted(self, event: FileSystemEvent):
        """Handle file deletion."""
        if event.is_directory or self._should_ignore(event.src_path):
            return

        relative_path = self._get_relative_path(event.src_path)
        self._add_event(
            EventType.FILE_DELETED,
            relative_path,
            f"{relative_path} deleted"
        )


class RepositoryWatcher:
    """Watches multiple repositories for file system changes."""

    def __init__(self, activity_feed: ActivityFeed, stats: RepositoryStats):
        """Initialize repository watcher.

        Args:
            activity_feed: Activity feed to add events to
            stats: Stats tracker
        """
        if not WATCHDOG_AVAILABLE:
            raise ImportError("watchdog library not installed. Install with: pip install watchdog")

        self.activity_feed = activity_feed
        self.stats = stats
        self.observer = Observer()
        self.watched_repos: Dict[str, RepositoryEventHandler] = {}
        self.is_watching = False

    def add_repository(self, repo_path: str):
        """Add repository to watch list.

        Args:
            repo_path: Path to repository
        """
        if repo_path in self.watched_repos:
            return  # Already watching

        repo_name = Path(repo_path).name

        # Get the current event loop
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # No running loop, get the default one
            loop = asyncio.get_event_loop()

        handler = RepositoryEventHandler(
            repo_path,
            repo_name,
            self.activity_feed,
            self.stats,
            loop
        )

        # Watch repository (recursive)
        self.observer.schedule(handler, repo_path, recursive=True)
        self.watched_repos[repo_path] = handler

    def remove_repository(self, repo_path: str):
        """Remove repository from watch list.

        Args:
            repo_path: Path to repository
        """
        if repo_path in self.watched_repos:
            # Note: watchdog doesn't have a clean way to remove individual watches
            # We'd need to restart the observer
            del self.watched_repos[repo_path]

    def start(self):
        """Start watching repositories."""
        if not self.is_watching:
            debug_log(f"[WATCHER] Starting observer, watching {len(self.watched_repos)} repos:")
            for repo_path in self.watched_repos.keys():
                debug_log(f"[WATCHER]   - {Path(repo_path).name}")
            self.observer.start()
            self.is_watching = True
            debug_log(f"[WATCHER] Observer started successfully")

    def stop(self):
        """Stop watching repositories."""
        if self.is_watching:
            self.observer.stop()
            self.observer.join()
            self.is_watching = False

    def pause(self):
        """Pause watching (same as stop for now)."""
        self.stop()

    def resume(self):
        """Resume watching (same as start for now)."""
        self.start()

    def get_watch_count(self) -> int:
        """Get number of watched repositories."""
        return len(self.watched_repos)


class GitStatusPoller:
    """Polls git status periodically for changes."""

    def __init__(
        self,
        activity_feed: ActivityFeed,
        stats: RepositoryStats,
        poll_interval: int = 10
    ):
        """Initialize git status poller.

        Args:
            activity_feed: Activity feed
            stats: Stats tracker
            poll_interval: Seconds between polls
        """
        self.activity_feed = activity_feed
        self.stats = stats
        self.poll_interval = poll_interval
        self.repos: Set[str] = set()
        self.is_polling = False
        self._task: Optional[asyncio.Task] = None
        self._repo_states: Dict[str, Dict] = {}

    def add_repository(self, repo_path: str):
        """Add repository to poll."""
        self.repos.add(repo_path)

    def remove_repository(self, repo_path: str):
        """Remove repository from polling."""
        self.repos.discard(repo_path)

    async def _check_git_status(self, repo_path: str) -> Dict[str, any]:
        """Check git status for a repository.

        Args:
            repo_path: Path to repository

        Returns:
            Status dict
        """
        try:
            # Check if repo is dirty
            result = await asyncio.create_subprocess_exec(
                'git', 'status', '--porcelain',
                cwd=repo_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE
            )
            stdout, _ = await result.communicate()
            is_dirty = len(stdout.decode().strip()) > 0

            # Check for unpushed commits
            result = await asyncio.create_subprocess_exec(
                'git', 'log', '@{u}..HEAD', '--oneline',
                cwd=repo_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE
            )
            stdout, _ = await result.communicate()
            unpushed_count = len(stdout.decode().strip().split('\n')) if stdout.decode().strip() else 0

            return {
                "dirty": is_dirty,
                "unpushed": unpushed_count > 0,
                "unpushed_count": unpushed_count
            }

        except Exception:
            return {"dirty": False, "unpushed": False, "unpushed_count": 0}

    async def _poll_loop(self):
        """Main polling loop."""
        while self.is_polling:
            for repo_path in list(self.repos):
                status = await self._check_git_status(repo_path)
                prev_status = self._repo_states.get(repo_path, {})

                repo_name = Path(repo_path).name

                # Check for status changes
                if status["dirty"] and not prev_status.get("dirty"):
                    await self.stats.mark_dirty(repo_path)

                elif not status["dirty"] and prev_status.get("dirty"):
                    # Repo became clean - likely committed
                    await self.stats.mark_clean(repo_path)
                    await self.activity_feed.add_event(ActivityEvent(
                        timestamp=datetime.now(timezone.utc),
                        repo_path=repo_path,
                        repo_name=repo_name,
                        event_type=EventType.GIT_COMMIT,
                        message="Changes committed"
                    ))

                if status["unpushed"] and not prev_status.get("unpushed"):
                    await self.stats.mark_unpushed(repo_path)

                elif not status["unpushed"] and prev_status.get("unpushed"):
                    # Commits were pushed
                    await self.stats.mark_pushed(repo_path)
                    await self.activity_feed.add_event(ActivityEvent(
                        timestamp=datetime.now(timezone.utc),
                        repo_path=repo_path,
                        repo_name=repo_name,
                        event_type=EventType.GIT_PUSH,
                        message="Commits pushed"
                    ))

                self._repo_states[repo_path] = status

            # Wait before next poll
            await asyncio.sleep(self.poll_interval)

    def start(self):
        """Start polling."""
        if not self.is_polling:
            self.is_polling = True
            self._task = asyncio.create_task(self._poll_loop())

    def stop(self):
        """Stop polling."""
        if self.is_polling:
            self.is_polling = False
            if self._task:
                self._task.cancel()



"""
MCP Server implementation for repoindex.

Implements the Model Context Protocol (MCP) to expose repoindex functionality
to LLM tools like Claude Code.

Architecture:
    MCPServer uses service layer (RepositoryService, TagService, EventService)
    to provide consistent, well-abstracted access to repoindex functionality.
"""

import json
import logging
from typing import Dict, Any, List, Optional, Callable
from dataclasses import dataclass, field
from pathlib import Path
from datetime import datetime, timedelta

# Domain layer
from ..domain import Repository, Tag, TagSource, Event

# Service layer
from ..services import RepositoryService, TagService, EventService

# Infrastructure layer
from ..infra import GitClient, GitHubClient, FileStore

# Config
from ..config import load_config

# Events parsing
from ..events import parse_timespec

logger = logging.getLogger(__name__)


@dataclass
class Resource:
    """Represents an MCP resource (read-only data)."""
    uri: str
    name: str
    description: str
    mime_type: str = "application/json"


@dataclass
class Tool:
    """Represents an MCP tool (action)."""
    name: str
    description: str
    input_schema: Dict[str, Any]
    handler: Callable


@dataclass
class MCPContext:
    """
    Shared context for MCP handlers.

    Holds service instances and configuration to avoid repetitive
    initialization in each handler.
    """
    config: Dict[str, Any]
    repo_service: RepositoryService
    tag_service: TagService
    event_service: EventService

    @classmethod
    def create(cls) -> 'MCPContext':
        """Create context with default services."""
        config = load_config()
        config_path = Path("~/.repoindex/config.json").expanduser()

        # Infrastructure
        git_client = GitClient()
        github_client = GitHubClient()
        file_store = FileStore(config_path)

        # Services
        repo_service = RepositoryService(
            config=config,
            git_client=git_client,
            github_client=github_client
        )
        tag_service = TagService(config_store=file_store)
        event_service = EventService(git_client=git_client)

        return cls(
            config=config,
            repo_service=repo_service,
            tag_service=tag_service,
            event_service=event_service
        )

    def discover_repos(self) -> List[Repository]:
        """Discover all repositories from config."""
        return list(self.repo_service.discover())

    def find_repo_by_name(self, name: str) -> Optional[Repository]:
        """Find a repository by name."""
        for repo in self.repo_service.discover():
            if repo.name == name:
                return repo
        return None


@dataclass
class MCPServer:
    """
    repoindex MCP Server.

    Exposes repoindex functionality as MCP resources and tools.
    """
    resources: Dict[str, Resource] = field(default_factory=dict)
    tools: Dict[str, Tool] = field(default_factory=dict)
    resource_handlers: Dict[str, Callable] = field(default_factory=dict)

    def register_resource(self, uri_pattern: str, name: str, description: str,
                         handler: Callable, mime_type: str = "application/json"):
        """Register a resource with its handler."""
        resource = Resource(
            uri=uri_pattern,
            name=name,
            description=description,
            mime_type=mime_type
        )
        self.resources[uri_pattern] = resource
        self.resource_handlers[uri_pattern] = handler
        logger.debug(f"Registered resource: {uri_pattern}")

    def register_tool(self, name: str, description: str,
                     input_schema: Dict[str, Any], handler: Callable):
        """Register a tool with its handler."""
        tool = Tool(
            name=name,
            description=description,
            input_schema=input_schema,
            handler=handler
        )
        self.tools[name] = tool
        logger.debug(f"Registered tool: {name}")

    def list_resources(self) -> List[Dict[str, Any]]:
        """List all available resources."""
        return [
            {
                "uri": r.uri,
                "name": r.name,
                "description": r.description,
                "mimeType": r.mime_type
            }
            for r in self.resources.values()
        ]

    def list_tools(self) -> List[Dict[str, Any]]:
        """List all available tools."""
        return [
            {
                "name": t.name,
                "description": t.description,
                "inputSchema": t.input_schema
            }
            for t in self.tools.values()
        ]

    def read_resource(self, uri: str) -> Dict[str, Any]:
        """Read a resource by URI."""
        # Try exact match first
        if uri in self.resource_handlers:
            return self.resource_handlers[uri]()

        # Try pattern matching
        for pattern, handler in self.resource_handlers.items():
            if self._match_uri_pattern(pattern, uri):
                params = self._extract_uri_params(pattern, uri)
                return handler(**params)

        raise ValueError(f"Unknown resource: {uri}")

    def call_tool(self, name: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Call a tool by name with arguments."""
        if name not in self.tools:
            raise ValueError(f"Unknown tool: {name}")

        tool = self.tools[name]
        return tool.handler(**arguments)

    def _match_uri_pattern(self, pattern: str, uri: str) -> bool:
        """Check if a URI matches a pattern with {param} placeholders."""
        import re
        # Convert pattern to regex
        regex_pattern = re.sub(r'\{(\w+)\}', r'([^/]+)', pattern)
        regex_pattern = f"^{regex_pattern}$"
        return bool(re.match(regex_pattern, uri))

    def _extract_uri_params(self, pattern: str, uri: str) -> Dict[str, str]:
        """Extract parameters from a URI given a pattern."""
        import re
        # Find all parameter names
        param_names = re.findall(r'\{(\w+)\}', pattern)

        # Convert pattern to regex with capture groups
        regex_pattern = re.sub(r'\{(\w+)\}', r'([^/]+)', pattern)
        regex_pattern = f"^{regex_pattern}$"

        match = re.match(regex_pattern, uri)
        if not match:
            return {}

        return dict(zip(param_names, match.groups()))


def create_mcp_server() -> MCPServer:
    """
    Create and configure the repoindex MCP server.

    Returns:
        Configured MCPServer instance
    """
    server = MCPServer()
    ctx = MCPContext.create()

    # === RESOURCE HANDLERS ===

    def list_repos():
        """List all repositories with basic metadata."""
        repos = ctx.discover_repos()
        result = []

        for repo in repos:
            # Get tags via service
            tags = ctx.tag_service.get_tag_strings(repo)

            result.append({
                'path': repo.path,
                'name': repo.name,
                'language': repo.language,
                'tags': tags,
                'has_package': repo.package is not None and repo.package.published
            })

        return {'repositories': result, 'count': len(result)}

    def get_repo(name: str):
        """Get full metadata for one repository."""
        repo = ctx.find_repo_by_name(name)

        if not repo:
            return {'error': f'Repository not found: {name}'}

        # Enrich with status and tags
        repo = ctx.repo_service.get_status(repo)
        tags = ctx.tag_service.get_tag_strings(repo)

        result = repo.to_dict()
        result['tags'] = tags
        return result

    def get_repo_status(name: str):
        """Get git status for one repository."""
        repo = ctx.find_repo_by_name(name)

        if not repo:
            return {'error': f'Repository not found: {name}'}

        repo = ctx.repo_service.get_status(repo)
        return {
            'name': name,
            'path': repo.path,
            'status': repo.status.to_dict()
        }

    def get_repo_package(name: str):
        """Get package info for one repository."""
        repo = ctx.find_repo_by_name(name)

        if not repo:
            return {'error': f'Repository not found: {name}'}

        return {
            'name': name,
            'path': repo.path,
            'package': repo.package.to_dict() if repo.package else {}
        }

    def list_tags():
        """List all tags."""
        repos = ctx.discover_repos()
        all_tags = ctx.tag_service.get_unique_tags(repos)
        return {'tags': sorted(list(all_tags)), 'count': len(all_tags)}

    def get_tag_tree():
        """Get hierarchical tag view."""
        repos = ctx.discover_repos()
        all_tags = ctx.tag_service.get_unique_tags(repos)

        # Build tree structure
        tree: Dict[str, Any] = {}
        for tag_str in sorted(all_tags):
            parts = tag_str.replace(':', '/').split('/')
            node = tree
            for part in parts:
                if part not in node:
                    node[part] = {}
                node = node[part]

        return {'tree': tree}

    def get_repos_by_tag(tag: str):
        """Get repositories with a specific tag."""
        repos = ctx.discover_repos()
        matching = []

        for repo in ctx.tag_service.query(repos, tag):
            tags = ctx.tag_service.get_tag_strings(repo)
            matching.append({
                'path': repo.path,
                'name': repo.name,
                'tags': tags
            })

        # Also match prefix patterns (tag/*)
        if not matching:
            for repo in ctx.tag_service.query(repos, f"{tag}/*"):
                tags = ctx.tag_service.get_tag_strings(repo)
                matching.append({
                    'path': repo.path,
                    'name': repo.name,
                    'tags': tags
                })

        return {'tag': tag, 'repositories': matching, 'count': len(matching)}

    def get_stats_summary():
        """Get overall statistics."""
        repos = ctx.discover_repos()

        languages: Dict[str, int] = {}
        published = 0
        total = len(repos)

        for repo in repos:
            lang = repo.language or 'Unknown'
            languages[lang] = languages.get(lang, 0) + 1
            if repo.package and repo.package.published:
                published += 1

        return {
            'total_repositories': total,
            'languages': languages,
            'published_packages': published,
            'unpublished_packages': total - published
        }

    def get_stats_languages():
        """Get repository count by language."""
        repos = ctx.discover_repos()

        languages: Dict[str, int] = {}
        for repo in repos:
            lang = repo.language or 'Unknown'
            languages[lang] = languages.get(lang, 0) + 1

        return {'languages': languages}

    def get_stats_published():
        """Get registry publication status."""
        repos = ctx.discover_repos()

        by_registry: Dict[str, int] = {'pypi': 0, 'cran': 0, 'npm': 0, 'unpublished': 0}

        for repo in repos:
            if repo.package and repo.package.published:
                registry = repo.package.registry or 'unknown'
                by_registry[registry] = by_registry.get(registry, 0) + 1
            else:
                by_registry['unpublished'] += 1

        return {'by_registry': by_registry}

    def get_recent_events_resource():
        """Get recent events (last 7 days)."""
        try:
            repos = ctx.discover_repos()
            events = ctx.event_service.get_recent(repos, days=7, limit=20)
            return {
                'events': [e.to_dict() for e in events],
                'count': len(events)
            }
        except Exception as e:
            logger.warning(f"Could not get events: {e}")
            return {'events': [], 'error': str(e)}

    def get_events_by_repo(name: str):
        """Get events for one repository."""
        try:
            repo = ctx.find_repo_by_name(name)

            if not repo:
                return {'error': f'Repository not found: {name}'}

            events = ctx.event_service.get_recent([repo], days=30, limit=50)
            return {
                'name': name,
                'events': [e.to_dict() for e in events],
                'count': len(events)
            }
        except Exception as e:
            return {'error': str(e)}

    def get_events_by_type(event_type: str):
        """Get events by type."""
        try:
            repos = ctx.discover_repos()
            events = list(ctx.event_service.scan(
                repos,
                types=[event_type],
                limit=50
            ))
            return {
                'type': event_type,
                'events': [e.to_dict() for e in events],
                'count': len(events)
            }
        except Exception as e:
            return {'error': str(e)}

    def get_events_since(timespec: str):
        """Get events since a time specification."""
        try:
            since = parse_timespec(timespec)
            repos = ctx.discover_repos()

            events = list(ctx.event_service.scan(
                repos,
                since=since,
                limit=100
            ))
            return {
                'since': timespec,
                'events': [e.to_dict() for e in events],
                'count': len(events)
            }
        except ValueError as e:
            return {'error': f'Invalid time specification: {e}'}
        except Exception as e:
            return {'error': str(e)}

    # === REGISTER RESOURCES ===

    server.register_resource(
        "repo://list",
        "Repository List",
        "List all repositories with basic metadata",
        list_repos
    )

    server.register_resource(
        "repo://{name}",
        "Repository Details",
        "Get full metadata for one repository",
        get_repo
    )

    server.register_resource(
        "repo://{name}/status",
        "Repository Status",
        "Get git status for one repository",
        get_repo_status
    )

    server.register_resource(
        "repo://{name}/package",
        "Repository Package",
        "Get package info (PyPI/CRAN/npm) for one repository",
        get_repo_package
    )

    server.register_resource(
        "tags://list",
        "Tags List",
        "List all tags",
        list_tags
    )

    server.register_resource(
        "tags://tree",
        "Tags Tree",
        "Get hierarchical tag view",
        get_tag_tree
    )

    server.register_resource(
        "tags://{tag}/repos",
        "Repositories by Tag",
        "Get repositories with a specific tag",
        get_repos_by_tag
    )

    server.register_resource(
        "stats://summary",
        "Statistics Summary",
        "Get overall statistics",
        get_stats_summary
    )

    server.register_resource(
        "stats://languages",
        "Language Statistics",
        "Get repository count by language",
        get_stats_languages
    )

    server.register_resource(
        "stats://published",
        "Publication Statistics",
        "Get registry publication status",
        get_stats_published
    )

    server.register_resource(
        "events://recent",
        "Recent Events",
        "Get recent events (last 7 days)",
        get_recent_events_resource
    )

    server.register_resource(
        "events://since/{timespec}",
        "Events Since",
        "Get events since a time (e.g., 1d, 7d, 2024-01-01)",
        get_events_since
    )

    server.register_resource(
        "events://repo/{name}",
        "Repository Events",
        "Get events for one repository",
        get_events_by_repo
    )

    server.register_resource(
        "events://type/{event_type}",
        "Events by Type",
        "Get events by type (git_tag, commit)",
        get_events_by_type
    )

    # === TOOL HANDLERS ===

    def tool_tag(repo: str, tag: str) -> Dict[str, Any]:
        """Add a tag to a repository."""
        repository = ctx.find_repo_by_name(repo)

        if not repository:
            return {'success': False, 'error': f'Repository not found: {repo}'}

        # Parse and validate tag
        tag_obj = Tag.parse(tag, TagSource.EXPLICIT)

        # Check if already has tag
        existing_tags = ctx.tag_service.get_explicit_tags(repository)
        if tag_obj in existing_tags:
            return {
                'success': True,
                'repo': repo,
                'tag': tag,
                'action': 'already_exists'
            }

        # Add tag
        ctx.tag_service.add(repository, tag_obj)

        return {
            'success': True,
            'repo': repo,
            'tag': tag,
            'action': 'added'
        }

    def tool_untag(repo: str, tag: str) -> Dict[str, Any]:
        """Remove a tag from a repository."""
        repository = ctx.find_repo_by_name(repo)

        if not repository:
            return {'success': False, 'error': f'Repository not found: {repo}'}

        # Parse tag
        tag_obj = Tag.parse(tag)

        # Remove tag
        removed = ctx.tag_service.remove(repository, tag_obj)

        return {
            'success': True,
            'repo': repo,
            'tag': tag,
            'action': 'removed' if removed else 'not_found'
        }

    def tool_query(expression: str) -> Dict[str, Any]:
        """Query repositories using the query language."""
        try:
            repos = ctx.discover_repos()
            results = []

            for repo in ctx.repo_service.filter_by_query(repos, expression):
                tags = ctx.tag_service.get_tag_strings(repo)
                result = repo.to_dict()
                result['tags'] = tags
                results.append(result)

            return {
                'query': expression,
                'results': results,
                'count': len(results)
            }
        except Exception as e:
            return {'query': expression, 'error': str(e)}

    def tool_refresh(repo: Optional[str] = None) -> Dict[str, Any]:
        """Refresh metadata for one or all repositories."""
        refreshed = []

        if repo:
            # Refresh single repo
            repository = ctx.find_repo_by_name(repo)
            if repository:
                ctx.repo_service.get_status(repository, fetch_github=True)
                refreshed.append(repo)
        else:
            # Refresh all repos
            for repository in ctx.discover_repos():
                ctx.repo_service.get_status(repository, fetch_github=True)
                refreshed.append(repository.name)

        return {
            'refreshed': refreshed,
            'count': len(refreshed)
        }

    def tool_stats(groupby: str = 'language') -> Dict[str, Any]:
        """Get statistics grouped by a field."""
        repos = ctx.discover_repos()

        groups: Dict[str, int] = {}
        for repo in repos:
            if groupby == 'language':
                key = repo.language or 'Unknown'
            elif groupby == 'published':
                key = 'published' if (repo.package and repo.package.published) else 'unpublished'
            elif groupby == 'registry':
                if repo.package and repo.package.published:
                    key = repo.package.registry or 'unknown'
                else:
                    key = 'none'
            else:
                # Try to get from repo dict
                repo_dict = repo.to_dict()
                key = str(repo_dict.get(groupby, 'Unknown'))

            groups[key] = groups.get(key, 0) + 1

        return {'groupby': groupby, 'groups': groups}

    # === REGISTER TOOLS ===

    server.register_tool(
        "repoindex_tag",
        "Add a tag to a repository",
        {
            "type": "object",
            "properties": {
                "repo": {"type": "string", "description": "Repository name"},
                "tag": {"type": "string", "description": "Tag to add (e.g., 'topic:ml')"}
            },
            "required": ["repo", "tag"]
        },
        tool_tag
    )

    server.register_tool(
        "repoindex_untag",
        "Remove a tag from a repository",
        {
            "type": "object",
            "properties": {
                "repo": {"type": "string", "description": "Repository name"},
                "tag": {"type": "string", "description": "Tag to remove"}
            },
            "required": ["repo", "tag"]
        },
        tool_untag
    )

    server.register_tool(
        "repoindex_query",
        "Query repositories using the repoindex query language",
        {
            "type": "object",
            "properties": {
                "expression": {
                    "type": "string",
                    "description": "Query expression (e.g., \"language == 'Python' and 'ml' in tags\")"
                }
            },
            "required": ["expression"]
        },
        tool_query
    )

    server.register_tool(
        "repoindex_refresh",
        "Refresh metadata for repositories",
        {
            "type": "object",
            "properties": {
                "repo": {
                    "type": "string",
                    "description": "Repository name (optional, refreshes all if not specified)"
                }
            },
            "required": []
        },
        tool_refresh
    )

    server.register_tool(
        "repoindex_stats",
        "Get repository statistics",
        {
            "type": "object",
            "properties": {
                "groupby": {
                    "type": "string",
                    "description": "Field to group by (language, published, registry)",
                    "default": "language"
                }
            },
            "required": []
        },
        tool_stats
    )

    return server


def run_mcp_server(transport: str = "stdio"):
    """
    Run the MCP server.

    Args:
        transport: Transport type ("stdio" or "http")
    """
    server = create_mcp_server()

    if transport == "stdio":
        _run_stdio_server(server)
    elif transport == "http":
        _run_http_server(server)
    else:
        raise ValueError(f"Unknown transport: {transport}")


def _run_stdio_server(server: MCPServer):
    """Run MCP server over stdio (JSON-RPC)."""
    import sys

    logger.info("Starting repoindex MCP server (stdio)")

    # Read JSON-RPC requests from stdin, write responses to stdout
    for line in sys.stdin:
        try:
            request = json.loads(line)
            response = _handle_jsonrpc_request(server, request)
            print(json.dumps(response), flush=True)
        except json.JSONDecodeError:
            error_response = {
                "jsonrpc": "2.0",
                "error": {"code": -32700, "message": "Parse error"},
                "id": None
            }
            print(json.dumps(error_response), flush=True)
        except Exception as e:
            logger.error(f"Error handling request: {e}")
            error_response = {
                "jsonrpc": "2.0",
                "error": {"code": -32603, "message": str(e)},
                "id": request.get("id") if 'request' in dir() else None
            }
            print(json.dumps(error_response), flush=True)


def _handle_jsonrpc_request(server: MCPServer, request: Dict[str, Any]) -> Dict[str, Any]:
    """Handle a JSON-RPC request."""
    method = request.get("method")
    params = request.get("params", {})
    request_id = request.get("id")

    result = None
    error = None

    try:
        if method == "initialize":
            result = {
                "protocolVersion": "2024-11-05",
                "capabilities": {
                    "resources": {"listChanged": True},
                    "tools": {}
                },
                "serverInfo": {
                    "name": "repoindex",
                    "version": "0.9.0"
                }
            }

        elif method == "resources/list":
            result = {"resources": server.list_resources()}

        elif method == "resources/read":
            uri = params.get("uri")
            content = server.read_resource(uri)
            result = {
                "contents": [{
                    "uri": uri,
                    "mimeType": "application/json",
                    "text": json.dumps(content, indent=2)
                }]
            }

        elif method == "tools/list":
            result = {"tools": server.list_tools()}

        elif method == "tools/call":
            tool_name = params.get("name")
            arguments = params.get("arguments", {})
            tool_result = server.call_tool(tool_name, arguments)
            result = {
                "content": [{
                    "type": "text",
                    "text": json.dumps(tool_result, indent=2)
                }]
            }

        else:
            error = {"code": -32601, "message": f"Method not found: {method}"}

    except Exception as e:
        error = {"code": -32603, "message": str(e)}

    response = {"jsonrpc": "2.0", "id": request_id}
    if error:
        response["error"] = error
    else:
        response["result"] = result

    return response


def _run_http_server(server: MCPServer, host: str = "localhost", port: int = 8765):
    """Run MCP server over HTTP (for testing/debugging)."""
    try:
        from http.server import HTTPServer, BaseHTTPRequestHandler

        class MCPHandler(BaseHTTPRequestHandler):
            def do_POST(self):
                content_length = int(self.headers['Content-Length'])
                body = self.rfile.read(content_length)

                try:
                    request = json.loads(body)
                    response = _handle_jsonrpc_request(server, request)

                    self.send_response(200)
                    self.send_header('Content-Type', 'application/json')
                    self.end_headers()
                    self.wfile.write(json.dumps(response).encode())

                except Exception as e:
                    self.send_response(500)
                    self.send_header('Content-Type', 'application/json')
                    self.end_headers()
                    error = {"jsonrpc": "2.0", "error": {"code": -32603, "message": str(e)}, "id": None}
                    self.wfile.write(json.dumps(error).encode())

            def log_message(self, format, *args):
                logger.debug(format % args)

        httpd = HTTPServer((host, port), MCPHandler)
        logger.info(f"Starting repoindex MCP server (HTTP) on {host}:{port}")
        httpd.serve_forever()

    except ImportError:
        raise RuntimeError("HTTP server not available")
